@@ -6,23 +6,26 @@ __author__ = 'Jens-Kristian Krogager'
 __email__ = "krogager@iap.fr"
 __credits__ = ["Jens-Kristian Krogager"]
 
+from copy import copy
 import numpy as np
-import astropy.io.fits as pf
+from astropy.io import fits
 import matplotlib.pyplot as plt
 from matplotlib import ticker
 from scipy import ndimage, signal
 import os
 from os.path import exists, basename
 
-from pynot import alfosc
+from pynot.data import organizer as organizer
+from pynot.logging import Report
+from pynot import instrument
 from pynot.functions import mad, my_formatter, get_version_number
-from pynot.scired import trim_overscan
-
+from pynot.scired import trim_overscan, correct_raw_file
+from pynot import reports
 
 __version__ = get_version_number()
 
 
-def combine_bias_frames(bias_frames, output='', kappa=15, method='mean', overwrite=True, overscan=50, mode='spec'):
+def combine_bias_frames(bias_frames, output='', kappa=15, method='mean', overwrite=True, mode='spec', report_fname=''):
     """Combine individual bias frames to create a 'master bias' frame.
     The combination is performed using robust sigma-clipping and
     median combination. Bad pixels are subsequently replaced by the
@@ -46,6 +49,9 @@ def combine_bias_frames(bias_frames, output='', kappa=15, method='mean', overwri
     method : string [default='mean']
         Method used for image combination (median/mean)
 
+    report_fname : string  [default='']
+        Filename of pdf diagnostic report
+
     Returns
     =======
     output : string
@@ -59,13 +65,10 @@ def combine_bias_frames(bias_frames, output='', kappa=15, method='mean', overwri
     bias = list()
     for frame in bias_frames:
         msg.append("          - Loaded bias frame: %s" % frame)
-        raw_img = pf.getdata(frame)
-        bias_hdr = alfosc.get_alfosc_header(frame)
-        if mode == 'spec':
-            trim_bias, bias_hdr = trim_overscan(raw_img, bias_hdr, overscan)
-            msg.append("          - Trimming overscan of bias images: %i!" % overscan)
-        else:
-            trim_bias = raw_img
+        raw_img = fits.getdata(frame)
+        bias_hdr = instrument.get_header(frame)
+        trim_bias, bias_hdr = trim_overscan(raw_img, bias_hdr)
+        msg.append("          - Trimming overscan of bias images")
         if len(bias) > 1:
             assert trim_bias.shape == bias[0].shape, "Images must have same shape!"
         bias.append(trim_bias)
@@ -90,11 +93,10 @@ def combine_bias_frames(bias_frames, output='', kappa=15, method='mean', overwri
         Ncomb = len(bias) - mask
         master_bias[Ncomb == 0] = np.mean(master_bias[Ncomb != 0])
     msg.append("          - Combined %i files" % len(bias))
+    msg.append("          - Image Stats:")
+    msg.append("          - Median = %.1f,  Std.Dev = %.1f" % (np.median(master_bias), 1.48*mad(master_bias)))
 
-    hdr = pf.getheader(bias_frames[0], 0)
-    hdr1 = pf.getheader(bias_frames[0], 1)
-    for key in hdr1.keys():
-        hdr[key] = hdr1[key]
+    hdr = bias_hdr
     hdr['NCOMBINE'] = len(bias_frames)
     if method.lower() == 'median':
         hdr.add_comment('Median combined Master Bias')
@@ -104,17 +106,21 @@ def combine_bias_frames(bias_frames, output='', kappa=15, method='mean', overwri
     if not output:
         output = 'MASTER_BIAS.fits'
 
-    pf.writeto(output, master_bias, header=hdr, overwrite=overwrite)
+    fits.writeto(output, master_bias, header=hdr, overwrite=overwrite)
     msg.append(" [OUTPUT] - Saving combined Bias Image: %s" % output)
+
+    if report_fname:
+        reports.check_bias(bias, bias_frames, output, report_fname=report_fname)
+        msg.append(" [OUTPUT] - Saving Bias Report: %s" % report_fname)
     msg.append("")
     output_msg = "\n".join(msg)
 
     return output, output_msg
 
 
-def combine_flat_frames(raw_frames, output, mbias='', mode='spec', dispaxis=2,
-                        kappa=5, verbose=False, overwrite=True, overscan=50,
-                        method='mean'):
+def combine_flat_frames(raw_frames, output, mbias='', mode='spec', dispaxis=None,
+                        kappa=5, verbose=False, overwrite=True, method='mean', report_fname='',
+                        **kwargs):
     """Combine individual spectral flat frames to create a 'master flat' frame.
     The individual frames are normalized to the mode of the 1D collapsed spectral
     shape. Individual frames are clipped using a kappa-sigma-clipping on the mode
@@ -153,6 +159,9 @@ def combine_flat_frames(raw_frames, output, mbias='', mode='spec', dispaxis=2,
     overwrite : boolean  [default=True]
         Overwrite existing output file if True.
 
+    report_fname : string  [default='']
+        Filename of pdf diagnostic report
+
     Returns
     =======
     output : string
@@ -163,28 +172,34 @@ def combine_flat_frames(raw_frames, output, mbias='', mode='spec', dispaxis=2,
     """
     msg = list()
     if mbias and exists(mbias):
-        bias = pf.getdata(mbias)
-        bias_hdr = alfosc.get_alfosc_header(mbias)
-        if mode == 'spec':
-            bias, bias_hdr = trim_overscan(bias, bias_hdr, overscan)
-            msg.append("          - Trimming overscan of bias image: %i!" % overscan)
+        bias = fits.getdata(mbias)
+        # bias_hdr = instrument.get_header(mbias)
+
     else:
         msg.append("[WARNING] - No master bias frame provided!")
         bias = 0.
 
+    if dispaxis is None:
+        hdr = instrument.get_header(raw_frames[0])
+        dispaxis = instrument.get_dispaxis(hdr)
+        msg.append("          - Getting dispersion axis from file: %i" % dispaxis)
+    elif dispaxis not in [1, 2]:
+        raise ValueError("dispaxis must be either 1 or 2, not: %r" % dispaxis)
+
     flats = list()
     flat_peaks = list()
     for fname in raw_frames:
-        hdr = alfosc.get_alfosc_header(fname)
-        flat = pf.getdata(fname)
+        hdr = instrument.get_header(fname)
+        flat = fits.getdata(fname)
+        flat, hdr = trim_overscan(flat, hdr)
+        msg.append("          - Trimming overscan of Flat images")
+
         if mode == 'spec':
-            flat, hdr = trim_overscan(flat, hdr, overscan)
             flat = flat - bias
             peak_val = np.max(np.mean(flat, dispaxis-1))
             flats.append(flat/peak_val)
             flat_peaks.append(peak_val)
             msg.append("          - Loaded Spectral Flat file: %s   mode=%.1f" % (fname, peak_val))
-            msg.append("          - Trimming overscan of Flat images: %i!" % overscan)
 
         else:
             flat = flat - bias
@@ -224,10 +239,7 @@ def combine_flat_frames(raw_frames, output, mbias='', mode='spec', dispaxis=2,
     else:
         msg.append("          - Combined %i files" % len(flats))
 
-    hdr = pf.getheader(raw_frames[0], 0)
-    hdr1 = pf.getheader(raw_frames[0], 1)
-    for key in hdr1.keys():
-        hdr[key] = hdr1[key]
+    # hdr = instrument.get_header(raw_frames[0])
     hdr['NCOMBINE'] = len(flats)
     if method.lower() == 'median':
         hdr.add_comment('Median combined Flat')
@@ -237,18 +249,15 @@ def combine_flat_frames(raw_frames, output, mbias='', mode='spec', dispaxis=2,
 
     if output == '':
         if mode == 'spec':
-            grism = alfosc.grism_translate[hdr['ALGRNM']]
-            output = 'flatcombine_%s.fits' % grism
+            grism = instrument.get_grism(hdr)
+            slit_name = instrument.get_slit(hdr)
+            output = 'flatcombine_%s_%s.fits' % (grism, slit_name)
         else:
             filter = 'white'
-            for keyword in ['FAFLTNM', 'FBFLTNM', 'ALFLTNM']:
-                if 'open' in hdr[keyword].lower():
-                    pass
-                else:
-                    filter = hdr[keyword]
+            filter = instrument.get_filter(hdr)
             output = 'flatcombine_%s.fits' % filter
 
-    pf.writeto(output, flat_combine, header=hdr, overwrite=overwrite)
+    fits.writeto(output, flat_combine, header=hdr, overwrite=overwrite)
     msg.append(" [OUTPUT] - Saving combined Flat Field Image: %s" % output)
     msg.append("")
     output_msg = "\n".join(msg)
@@ -258,7 +267,7 @@ def combine_flat_frames(raw_frames, output, mbias='', mode='spec', dispaxis=2,
     return output, output_msg
 
 
-def detect_flat_edges(img, dispaxis=2, savgol_window=21, threshold=10):
+def detect_flat_edges(img, dispaxis=2, savgol_window=21, threshold=10, width=10):
     """Use filtered derivative to detect the edges of the frame."""
     # Get median profile along slit:
     med_row = np.nanmedian(img, 2-dispaxis)
@@ -271,14 +280,14 @@ def detect_flat_edges(img, dispaxis=2, savgol_window=21, threshold=10):
     if noise == 0:
         sig_tmp = np.std(deriv)
         noise = np.std(deriv[deriv < 3*sig_tmp])
-    edges, props = signal.find_peaks(deriv, height=threshold*noise)
+    edges, props = signal.find_peaks(deriv, height=threshold*noise, width=width)
 
     return edges
 
 
 
-def normalize_spectral_flat(fname, output='', fig_dir='', dispaxis=2, overscan=50, order=24, savgol_window=51,
-                            med_window=5, edge_threshold=10, edge_window=21, plot=True, overwrite=True, **kwargs):
+def normalize_spectral_flat(fname, output='', fig_dir='', dispaxis=None, order=24, savgol_window=51,
+                            med_window=5, edge_threshold=10, edge_window=21, edge_width=10, plot=True, overwrite=True, **kwargs):
     """
     Normalize spectral flat field for long-slit observations. Parameters are optimized
     for NOT/ALFOSC spectra with horizontal slits, i.e., vertical spectra [axis=2],
@@ -295,7 +304,8 @@ def normalize_spectral_flat(fname, output='', fig_dir='', dispaxis=2, overscan=5
         Input FITS file with raw lamp flat data
 
     output : string  [default='']
-        Filename of normalized flat frame, if not given the output is not saved to file
+        Filename of normalized flat frame.
+        If not given, the output is constructed from the grism and slit IDs
 
     fig_dir : string  [default='']
         Directory where the diagnostic plot is saved (if `plot=True`)
@@ -303,9 +313,6 @@ def normalize_spectral_flat(fname, output='', fig_dir='', dispaxis=2, overscan=5
     dispaxis : integer  [default=2]
         Dispersion axis, 1: horizontal spectra, 2: vertical spectra
 
-    overscan : integer  [default=50]
-        Overscan region, default for ALFOSC is 50 pixels on either side and on top
-
     order : integer  [default=24]
         Order for Chebyshev polynomial to fit to the spatial profile (per row/col)
 
@@ -321,6 +328,10 @@ def normalize_spectral_flat(fname, output='', fig_dir='', dispaxis=2, overscan=5
     edge_window : integer  [default=21]
         The Savitzky--Golay window used for automatic edge detection
 
+    edge_width : integer  [default=10]
+        The minimum width of a peak in the 1st derivative of the filtered
+        illumination profile along the slit in units of pixels.
+
     plot : boolean [default=True]
         Plot the 1d and 2d data for inspection?
 
@@ -337,13 +348,20 @@ def normalize_spectral_flat(fname, output='', fig_dir='', dispaxis=2, overscan=5
 
     """
     msg = list()
-    flat = pf.getdata(fname)
-    hdr = alfosc.get_alfosc_header(fname)
-
+    flat = fits.getdata(fname)
+    hdr = instrument.get_header(fname)
+    grism = instrument.get_grism(hdr)
+    slit_name = instrument.get_slit(hdr)
     msg.append("          - Input file: %s" % fname)
+    msg.append("          - Grism Name: %s" % grism)
+    msg.append("          - Slit Name: %s" % slit_name)
+
+    if dispaxis is None:
+        dispaxis = instrument.get_dispaxis(hdr)
+        msg.append("          - Getting dispersion axis from file: %i" % dispaxis)
+    elif dispaxis not in [1, 2]:
+        raise ValueError("dispaxis must be either 1 or 2, not: %r" % dispaxis)
 
-    flat, hdr = trim_overscan(flat, hdr, overscan)
-    msg.append("          - Trimmed overscan: %i" % overscan)
     # Get raw pixel array of spatial axis
     x = np.arange(flat.shape[dispaxis-1])
 
@@ -376,13 +394,16 @@ def normalize_spectral_flat(fname, output='', fig_dir='', dispaxis=2, overscan=5
     msg.append("          - Fitting each spatial row/column using Chebyshev polynomials combined with Savitzky--Golay filtering")
     msg.append("          - Polynomial order: %i" % order)
     msg.append("          - Savitzky--Golay filter width: %i" % savgol_window)
-    pad = savgol_window // 2
+    pad = savgol_window // 2 + 1
     x_fit = x[x1:x2]
     for num, row in enumerate(smoothed_flat):
-        filtered_row = signal.savgol_filter(row[x1:x2], savgol_window, 1)
+        filtered_row = signal.savgol_filter(row[x1:x2], savgol_window, 2)
+        sig = 1.5*mad(row[x1:x2] - filtered_row)
+        # mask = (row[x1:x2] - filtered_row) > -sig
+        mask = np.abs(row[x1:x2] - filtered_row) < 2*sig
         # Exclude filter edges, half filter width:
-        x_mask = x_fit[pad:-pad]
-        row_mask = filtered_row[pad:-pad]
+        x_mask = x_fit[mask][pad:-pad]
+        row_mask = filtered_row[mask][pad:-pad]
         fit_row = np.polynomial.Chebyshev.fit(x_mask, row_mask, order, domain=[x_fit.min(), x_fit.max()])
         model1d = fit_row(x_fit)
         # Remove edge effects in fitting, the filtered data are more robust:
@@ -397,13 +418,14 @@ def normalize_spectral_flat(fname, output='', fig_dir='', dispaxis=2, overscan=5
         model = model.T
 
     flat_norm = flat / model
-    hdr['DATAMIN'] = np.min(flat_norm)
-    hdr['DATAMAX'] = np.max(flat_norm)
+    hdr.remove('DATAMIN', ignore_missing=True)
+    hdr.remove('DATAMAX', ignore_missing=True)
     if dispaxis == 1:
         stat_region = flat_norm[x1:x2, :]
     else:
         stat_region = flat_norm[:, x1:x2]
-    noise = np.std(stat_region)
+    # noise = np.std(stat_region)
+    noise = mad(stat_region)*1.48
     data_range = (np.min(stat_region), np.max(stat_region), np.median(stat_region))
     msg.append("          - Standard deviation of 1D residuals: %.2f ADUs" % noise)
     msg.append("          - Normalized data range: min=%.2e  max=%.2e  median=%.2e" % data_range)
@@ -416,11 +438,11 @@ def normalize_spectral_flat(fname, output='', fig_dir='', dispaxis=2, overscan=5
         ax1_2d = fig2D.add_subplot(121)
         ax2_2d = fig2D.add_subplot(122)
         ax1_2d.imshow(flat, origin='lower')
-        ax1_2d.set_title("Raw Flat")
+        ax1_2d.set_title("Combined Flat\n(%s - %s)" % (grism, slit_name))
         v1 = data_range[2] - 3*noise
         v2 = data_range[2] + 3*noise
         ax2_2d.imshow(flat_norm, origin='lower', vmin=v1, vmax=v2)
-        ax2_2d.set_title("Normalized Flat")
+        ax2_2d.set_title("Normalized Flat\n(%s - %s)" % (grism, slit_name))
         ax2_2d.set_yticklabels("")
         if dispaxis == 2:
             ax1_2d.set_xlabel("Spatial Axis [pixels]", fontsize=11)
@@ -436,9 +458,11 @@ def normalize_spectral_flat(fname, output='', fig_dir='', dispaxis=2, overscan=5
         ax2_1d = fig1D.add_subplot(212)
 
         flat1D = np.nanmedian(flat, 2-dispaxis)
-        f1d = signal.savgol_filter(flat1D[x1:x2], savgol_window, 1)
-        x_mask = x_fit[pad:-pad]
-        row_mask = f1d[pad:-pad]
+        f1d = signal.savgol_filter(flat1D[x1:x2], savgol_window, 2)
+        sig1d = 1.5*mad(flat1D[x1:x2] - f1d)
+        mask = np.abs(flat1D[x1:x2] - f1d) < 2*sig1d
+        x_mask = x_fit[mask][pad:-pad]
+        row_mask = f1d[mask][pad:-pad]
         fit_row = np.polynomial.Chebyshev.fit(x_mask, row_mask, order, domain=[x_fit.min(), x_fit.max()])
         flat_model = fit_row(x_fit)
         # Remove edge effects in fitting, the filtered data are more robust:
@@ -462,10 +486,11 @@ def normalize_spectral_flat(fname, output='', fig_dir='', dispaxis=2, overscan=5
         ax1_1d.set_ylabel('Counts  [$10^{{{0:d}}}$ ADU]'.format(int(power)), fontsize=11)
 
         ax2_1d.set_ylabel('($F_{\\rm 1D} -$ model) / model', fontsize=11)
-        plot_noise = np.nanstd(residuals/flat_model)
-        ax2_1d.set_ylim(-5*plot_noise, 5*plot_noise)
+        plot_noise = 1.48*mad(residuals/flat_model)
+        ax2_1d.set_ylim(-10*plot_noise, 10*plot_noise)
         ax2_1d.set_xlim(x.min(), x.max())
         ax1_1d.set_xlim(x.min(), x.max())
+        ax1_1d.set_ylim(flat_model.min()-10*sig1d, flat_model.max()+10*sig1d)
 
         ax1_1d.minorticks_on()
         ax2_1d.minorticks_on()
@@ -473,7 +498,7 @@ def normalize_spectral_flat(fname, output='', fig_dir='', dispaxis=2, overscan=5
         if not exists(fig_dir) and fig_dir != '':
             os.mkdir(fig_dir)
         file_base = basename(fname)
-        fname_root = file_base.strip('.fits')
+        fname_root = os.path.splitext(file_base)[0]
         fig1d_fname = os.path.join(fig_dir, "specflat_1d_%s.pdf" % fname_root)
         fig2d_fname = os.path.join(fig_dir, "specflat_2d_%s.pdf" % fname_root)
         fig1D.tight_layout()
@@ -495,13 +520,120 @@ def normalize_spectral_flat(fname, output='', fig_dir='', dispaxis=2, overscan=5
         else:
             output += '.fits'
     else:
-        grism = alfosc.grism_translate[hdr['ALGRNM']]
-        slit_name = hdr['ALAPRTNM']
         output = 'NORM_FLAT_%s_%s.fits' % (grism, slit_name)
 
-    pf.writeto(output, flat_norm, header=hdr, overwrite=overwrite)
+    fits.writeto(output, flat_norm, header=hdr, overwrite=overwrite)
     msg.append(" [OUTPUT] - Saving normalized MASTER FLAT: %s" % output)
     msg.append("")
     output_msg = "\n".join(msg)
 
     return output, output_msg
+
+
+def task_bias(options, database, log=None, verbose=True, output_dir='', report_dir=reports.report_dir):
+    """
+    Define the entry point for the task pynot:bias. This will automatcally
+    """
+    if log is None:
+        log = Report(verbose)
+    log.write("Running task: Bias combination")
+
+    if not os.path.exists(report_dir):
+        os.makedirs(report_dir)
+
+    if not os.path.exists(output_dir):
+        os.makedirs(output_dir)
+
+    bias_files = organizer.sort_bias(database['BIAS'])
+
+    tag = 'MBIAS'
+    task_output = {tag: []}
+    for file_id, input_list in bias_files.items():
+        output_fname = os.path.join(output_dir, '%s_%s.fits' % (tag, file_id))
+        report_fname = os.path.join(report_dir, '%s_%s_report.pdf' % (tag, file_id))
+        _, output_msg = combine_bias_frames(input_list, output_fname, report_fname=report_fname,
+                                            **options)
+        task_output[tag].append(output_fname)
+        log.commit(output_msg)
+        log.add_linebreak()
+    log.add_linebreak()
+    return task_output, log
+
+
+def task_sflat(options, database, log=None, verbose=True, output_dir='', report_dir=reports.report_dir):
+    """
+    Define the entry point for the main task of sflat, to be called by pynot.main
+
+    args : command line arguments from argparse of pynot.main
+    """
+    if log is None:
+        log = Report(verbose)
+    log.write("Running task: Spectral flat field combination and normalization")
+
+    flat_files = organizer.sort_spec_flat(database['SPEC_FLAT'])
+
+    tag = 'NORM_SFLAT'
+    task_output = {tag: []}
+    for file_id, input_list in flat_files.items():
+        # Match master bias file:
+        raw_img = organizer.RawImage(input_list[0])
+        master_bias = organizer.match_single_calib(raw_img, database, 'MBIAS', log, date=False)
+
+        output_fname = os.path.join(output_dir, 'FLAT_COMBINED_%s.fits' % file_id)
+        flatcombine, msg = combine_flat_frames(input_list, output=output_fname, mbias=master_bias,
+                                               mode='spec', **options)
+        log.commit(msg)
+        log.add_linebreak()
+
+        output_fname = os.path.join(output_dir, '%s_%s.fits' % (tag, file_id))
+        _, flat_msg = normalize_spectral_flat(flatcombine, output=output_fname,
+                                              fig_dir=output_dir, **options)
+        log.commit(flat_msg)
+        log.add_linebreak()
+        task_output[tag].append(output_fname)
+    return task_output, log
+
+
+def task_prep_arcs(options, database, log=None, verbose=True, output_dir='', report_dir=reports.report_dir):
+    """
+    Prepare all arc frames for analysis: apply bias and flat field corrections
+    """
+    if log is None:
+        log = Report(verbose)
+    log.write("Running task: Arc frame preparation")
+
+    arc_filelist = []
+    for tag in database.keys():
+        if 'ARC' in tag and 'CORR' not in tag:
+            arc_filelist += database[tag]
+
+    if len(arc_filelist) == 0:
+        log.error("No arc line calibration data found in the dataset!")
+        log.error("Check the classification table... object type 'ARC_HeNe', 'ARC_ThAr', 'ARC_HeAr' missing")
+        raise KeyError('No ARC files found in database!')
+
+    arc_files = organizer.sort_arcs(arc_filelist)
+    tag = 'ARC_CORR'
+    task_output = {tag: []}
+    for file_id, input_list in arc_files.items():
+        # find BIAS and FLAT
+        raw_img = organizer.RawImage(input_list[0])
+        master_bias = organizer.match_single_calib(raw_img, database, 'MBIAS', log, date=False)
+        norm_flat = organizer.match_single_calib(raw_img, database, 'NORM_SFLAT', log, date=False,
+                                                 grism=True, slit=True, filter=True)
+        current_fnames = list()
+        for arc_fname in input_list:
+            arc_basename = 'CORR_' + os.path.basename(arc_fname)
+            corrected_arc2d_fname = os.path.join(output_dir, arc_basename)
+            output_msg = correct_raw_file(arc_fname, bias_fname=master_bias, flat_fname=norm_flat,
+                                          output=corrected_arc2d_fname, overwrite=True)
+            log.commit(output_msg)
+            task_output[tag].append(corrected_arc2d_fname)
+            current_fnames.append(corrected_arc2d_fname)
+        # Create PDF diagnostic report
+        report_fname = os.path.join(report_dir, 'ARCS_%s_report.pdf' % file_id)
+        reports.check_arcs(current_fnames, report_fname)
+        log.add_linebreak()
+
+
+    return task_output, log
