@@ -11,6 +11,7 @@ import warnings
 from lmfit import Parameters, minimize
 
 from pynot.functions import mad, NN_moffat, NN_gaussian, fix_nans, get_version_number
+from pynot import instrument
 
 __version__ = get_version_number()
 
@@ -111,7 +112,7 @@ def median_filter_data(x, kappa=5., window=21):
     return (med_x, mask)
 
 
-def fit_trace(img2D, x, y, model_name='moffat', dx=50, ymin=5, ymax=-5, xmin=None, xmax=None):
+def fit_trace(img2D, x, y, model_name='moffat', dx=50, kappa=10., ymin=5, ymax=-5, xmin=None, xmax=None):
     """
     Perform automatic localization of the trace if possible, otherwise use fixed
     aperture to extract the 1D spectrum.
@@ -138,7 +139,6 @@ def fit_trace(img2D, x, y, model_name='moffat', dx=50, ymin=5, ymax=-5, xmin=Non
     spsf[ymax:] = 0.
 
     # Detect peaks:
-    kappa = 10.
     noise = mad(spsf)*1.48
     peaks, properties = find_peaks(spsf, prominence=kappa*noise, width=3)
     prominences = properties['prominences']
@@ -182,7 +182,7 @@ def fit_trace(img2D, x, y, model_name='moffat', dx=50, ymin=5, ymax=-5, xmin=Non
     return (x_binned, N_obj, trace_parameters, fwhm, output_msg)
 
 
-def create_2d_profile(img2D, model_name='moffat', dx=25, width_scale=2,
+def create_2d_profile(img2D, model_name='moffat', dx=25, width_scale=2, kappa_det=10.,
                       xmin=None, xmax=None, ymin=None, ymax=None, order_center=3, order_width=0,
                       w_cen=15, kappa_cen=3., w_width=21, kappa_width=3.):
     """
@@ -259,7 +259,7 @@ def create_2d_profile(img2D, model_name='moffat', dx=25, width_scale=2,
         if fwhm is None:
             raise ValueError("FWHM of the spectral trace could not be determined! Maybe more than one object in slit...")
     else:
-        fit_values = fit_trace(img2D, x, y, model_name=model_name, dx=dx, ymin=ymin, ymax=ymax, xmin=xmin, xmax=xmax)
+        fit_values = fit_trace(img2D, x, y, model_name=model_name, dx=dx, ymin=ymin, ymax=ymax, xmin=xmin, xmax=xmax, kappa=kappa_det)
     x_binned, N_obj, trace_parameters, fwhm, fit_msg = fit_values
     msg.append(fit_msg)
 
@@ -271,7 +271,6 @@ def create_2d_profile(img2D, model_name='moffat', dx=25, width_scale=2,
     trace_info = list()
     domain = [0, img2D.shape[1]]
     for n in range(N_obj):
-        msg.append("          - Working on profile number %i" % (n+1))
         info_dict = dict()
         info_dict['x_binned'] = x_binned
         # Median filter
@@ -279,6 +278,7 @@ def create_2d_profile(img2D, model_name='moffat', dx=25, width_scale=2,
         mu_err = np.array([p['mu_%i' % n].stderr for p in trace_parameters])
         mu_err[mu_err == 0] = 100.
         w_mu = 1./mu_err**2
+        w_mu[np.isnan(w_mu)] = 0.
         mu_med, mask_mu = median_filter_data(mu, kappa_cen, w_cen)
         mask_mu &= (x_binned > xmin) & (x_binned < xmax)
         mu_fit = Chebyshev.fit(x_binned[mask_mu], mu[mask_mu], deg=order_center, domain=domain, w=w_mu[mask_mu])
@@ -325,8 +325,8 @@ def create_2d_profile(img2D, model_name='moffat', dx=25, width_scale=2,
             mask_a &= (x_binned > xmin) & (x_binned < xmax)
             mask_b &= (x_binned > xmin) & (x_binned < xmax)
 
-            a_fit = Chebyshev.fit(x_binned[mask_a], a[mask_a], deg=order_width, domain=domain, w=w_a[mask_a])
-            b_fit = Chebyshev.fit(x_binned[mask_b], b[mask_b], deg=order_width, domain=domain, w=w_b[mask_b])
+            a_fit = Chebyshev.fit(x_binned[mask_a], a[mask_a], deg=order_width, domain=domain)  # , w=w_a[mask_a])
+            b_fit = Chebyshev.fit(x_binned[mask_b], b[mask_b], deg=order_width, domain=domain)  # , w=w_b[mask_b])
             info_dict['a'] = a
             info_dict['a_err'] = a_err
             info_dict['mask_a'] = mask_a
@@ -400,11 +400,13 @@ def plot_diagnostics(pdf, spec1D, err1D, info_dict, width_scale=2):
 
     axes[-1].plot(spec1D, color='k', lw=1.0, alpha=0.9, label='Flux')
     axes[-1].plot(err1D, color='crimson', lw=0.7, alpha=0.8, label='Error')
-    ymin = 0.
-    good = spec1D > 5*err1D
-    if np.sum(good) == 0:
-        good = spec1D[100:-100] > 0
-    ymax = np.nanmax(spec1D[good])
+    good = np.ones_like(spec1D, dtype=bool)
+    good[:50] = False
+    good[-50:] = False
+    noise = mad(spec1D[good])*1.48
+    ymax = np.nanmax(median_filter(spec1D[good], 15)) + 5*noise
+    ymin = -2*noise
+    axes[-1].axhline(0., ls=':', color='0.5', lw=0.5)
     axes[-1].set_ylim(ymin, ymax)
     axes[-1].set_ylabel("Flux")
     axes[-1].set_xlabel("Dispersion Axis  [pixels]")
@@ -414,7 +416,10 @@ def plot_diagnostics(pdf, spec1D, err1D, info_dict, width_scale=2):
 
 
 
-def auto_extract_img(img2D, err2D, *, N=None, pdf_fname=None, mask=None, model_name='moffat', dx=50, width_scale=2, xmin=None, xmax=None, ymin=None, ymax=None, order_center=3, order_width=0, w_cen=15, kappa_cen=3., w_width=21, kappa_width=3.):
+def auto_extract_img(img2D, err2D, *, N=None, pdf_fname=None, mask=None, model_name='moffat',
+                     dx=50, width_scale=2, xmin=None, xmax=None, ymin=None, ymax=None,
+                     order_center=3, order_width=0, w_cen=15, kappa_cen=3., w_width=21, kappa_width=3.,
+                     kappa_det=10.):
     assert err2D.shape == img2D.shape, "input image and error image do not match in shape"
     if N == 0:
         raise ValueError("Invalid input: N must be an integer larger than or equal to 1, not %r" % N)
@@ -435,7 +440,8 @@ def auto_extract_img(img2D, err2D, *, N=None, pdf_fname=None, mask=None, model_n
     profile_values = create_2d_profile(img2D, model_name=model_name, dx=dx, width_scale=width_scale,
                                        xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax,
                                        order_center=order_center, order_width=order_width,
-                                       w_cen=w_cen, kappa_cen=kappa_cen, w_width=w_width, kappa_width=kappa_width)
+                                       w_cen=w_cen, kappa_cen=kappa_cen, w_width=w_width, kappa_width=kappa_width,
+                                       kappa_det=kappa_det)
     trace_models_2d, trace_info, profile_msg = profile_values
     msg.append(profile_msg)
 
@@ -454,11 +460,15 @@ def auto_extract_img(img2D, err2D, *, N=None, pdf_fname=None, mask=None, model_n
 
     spectra = list()
     for P, info_dict in zip(trace_models_2d, trace_info):
-        spec1D = np.sum(M*P*img2D/var2D, axis=0) / np.sum(M*P**2/var2D, axis=0)
-        var1D = np.sum(M*P, axis=0) / np.sum(M*P**2/var2D, axis=0)
-        err1D = np.sqrt(var1D)
-        err1D = fix_nans(err1D)
-        spectra.append([spec1D, err1D])
+        with warnings.catch_warnings():
+            warnings.simplefilter('ignore')
+            spec1D = np.sum(M*P*img2D/var2D, axis=0) / np.sum(M*P**2/var2D, axis=0)
+            var1D = np.sum(M*P, axis=0) / np.sum(M*P**2/var2D, axis=0)
+            err1D = np.sqrt(var1D)
+            err1D = fix_nans(err1D)
+            mask1D = np.sum((1-M)*P, axis=0) / np.sum((1-M)*P**2, axis=0) > 0
+            trace_pos = np.median(info_dict['fit_mu'])
+            spectra.append([spec1D, err1D, mask1D, trace_pos])
 
         if pdf_fname:
             plot_diagnostics(pdf, spec1D, err1D, info_dict, width_scale)
@@ -472,35 +482,53 @@ def auto_extract_img(img2D, err2D, *, N=None, pdf_fname=None, mask=None, model_n
     return spectra, output_msg
 
 
-def auto_extract(fname, output, dispaxis=1, *, N=None, pdf_fname=None, mask=None, model_name='moffat', dx=50, width_scale=2, xmin=None, xmax=None, ymin=None, ymax=None, order_center=3, order_width=1, w_cen=15, kappa_cen=3., w_width=21, kappa_width=3., **kwargs):
+def auto_extract(fname, output, dispaxis=1, *, N=None, pdf_fname=None, model_name='moffat',
+                 dx=50, width_scale=2, xmin=None, xmax=None, ymin=None, ymax=None,
+                 order_center=3, order_width=1, w_cen=15, kappa_cen=3., w_width=21, kappa_width=3., kappa_det=10., **kwargs):
     """Automatically extract object spectra in the given file. Dispersion along the x-axis is assumed!"""
     msg = list()
     img2D = fits.getdata(fname)
     hdr = fits.getheader(fname)
     if 'DISPAXIS' in hdr:
         dispaxis = hdr['DISPAXIS']
-
     msg.append("          - Loaded image data: %s" % fname)
-    try:
-        err2D = fits.getdata(fname, 'ERR')
-        msg.append("          - Loaded error image extension")
-    except:
+
+    mask2D = None
+    err2D = None
+    with fits.open(fname) as hdu_list:
+        for hdu in hdu_list:
+            if hdu.name.upper() in ['MASK', 'QUAL', 'QC', 'DQ']:
+                mask2D = hdu.data
+                msg.append("          - Loaded pixel mask")
+            elif hdu.name.upper() in ['ERR', 'ERROR', 'ERRS', 'FLUX_ERR', 'FLUX_ERRS']:
+                err2D = hdu.data
+                msg.append("          - Loaded error image extension")
+            elif hdu.name.upper() in ['IVAR']:
+                err2D = 1./np.sqrt(hdu.data)
+                msg.append("          - Loaded inverse variance image extension")
+                msg.append("          - Converted inverse variance to error image")
+
+    if err2D is None:
         noise = 1.5*mad(img2D)
         err2D = np.ones_like(img2D) * noise
         msg.append("[WARNING] - No error image detected!")
         msg.append("[WARNING] - Generating one from image statistics:")
         msg.append("[WARNING] - Median=%.2e  Sigma=%.2e" % (np.nanmedian(img2D), noise))
 
+    if mask2D is None:
+        mask2D = np.zeros_like(img2D)
 
     if dispaxis == 2:
         img2D = img2D.T
         err2D = err2D.T
+        mask2D = mask2D.T
 
-    spectra, ext_msg = auto_extract_img(img2D, err2D, N=N, pdf_fname=pdf_fname, mask=mask,
+    spectra, ext_msg = auto_extract_img(img2D, err2D, N=N, pdf_fname=pdf_fname, mask=mask2D,
                                         model_name=model_name, dx=dx, width_scale=width_scale,
                                         xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax,
                                         order_center=order_center, order_width=order_width,
-                                        w_cen=w_cen, kappa_cen=kappa_cen, w_width=w_width, kappa_width=kappa_width)
+                                        w_cen=w_cen, kappa_cen=kappa_cen, w_width=w_width, kappa_width=kappa_width,
+                                        kappa_det=kappa_det)
     msg.append(ext_msg)
 
     hdu = fits.HDUList()
@@ -516,20 +544,28 @@ def auto_extract(fname, output, dispaxis=1, *, N=None, pdf_fname=None, mask=None
         wl = np.arange(len(spectra[0][0]))
 
 
+    wl_unit = hdr.get('CUNIT1')
+    flux_unit = hdr.get('BUNIT')
     keywords_base = ['CDELT%i', 'CRPIX%i', 'CRVAL%i', 'CTYPE%i', 'CUNIT%i']
     keywords_to_remove = sum([[key % num for key in keywords_base] for num in [1, 2]], [])
     keywords_to_remove += ['CD1_1', 'CD2_1', 'CD1_2', 'CD2_2']
     keywords_to_remove += ['BUNIT', 'DATAMIN', 'DATAMAX']
-    for num, (flux, err) in enumerate(spectra):
-        col_wl = fits.Column(name='WAVE', array=wl, format='D', unit=hdr['CUNIT1'])
-        col_flux = fits.Column(name='FLUX', array=flux, format='D', unit=hdr['BUNIT'])
-        col_err = fits.Column(name='ERR', array=err, format='D', unit=hdr['BUNIT'])
+    for num, (flux, err, mask1d, trace_pos) in enumerate(spectra):
+        col_wl = fits.Column(name='WAVE', array=wl, format='D', unit=wl_unit)
+        col_flux = fits.Column(name='FLUX', array=flux, format='D', unit=flux_unit)
+        col_err = fits.Column(name='ERR', array=err, format='D', unit=flux_unit)
+        col_mask = fits.Column(name='MASK', array=mask1d, format='L')
+        hdr['OBJ_POS'] = (np.round(trace_pos, 2), "Median extraction position along slit [pixels]")
         for key in keywords_to_remove:
             hdr.remove(key, ignore_missing=True)
-        tab = fits.BinTableHDU.from_columns([col_wl, col_flux, col_err], header=hdr)
+        tab = fits.BinTableHDU.from_columns([col_wl, col_flux, col_err, col_mask], header=hdr)
         tab.name = 'OBJ%i' % (num+1)
         hdu.append(tab)
 
+    if not output:
+        target_name = instrument.get_target_name(hdr)
+        output = 'FLUX1D_%s.fits' % target_name
+
     hdu.writeto(output, overwrite=True, output_verify='silentfix')
     msg.append(" [OUTPUT] - Writing fits table: %s" % output)
     msg.append("")
