@@ -1,3 +1,4 @@
+import astropy.units as u
 import numpy as np
 import yaml
 import os
@@ -212,3 +213,100 @@ def decimal_to_string(ra, dec, delimiter=':'):
     dec_str = delimiter.join(dms)
 
     return (ra_str, dec_str)
+
+
+def get_pixtab_parameters(pixtable_fname):
+    """Find the polynomial degree, ref_type, loc used when creating the pixel table"""
+    found_order = False
+    order_wl = 4
+    found_ref = False
+    ref_type = 'vacuum'
+    found_loc = False
+    loc = -1
+    with open(pixtable_fname) as tab_file:
+        all_lines = tab_file.readlines()
+
+    for line in all_lines:
+        if line[0] != '#':
+            # Reached the end of the header
+            break
+        elif 'order' in line:
+            order_str = line.split('=')[1]
+            order_wl = int(order_str.strip())
+            found_order = True
+        elif 'ref' in line:
+            ref_str = line.split('=')[1]
+            ref_type = ref_str.strip()
+            found_ref = True
+        elif 'loc' in line:
+            loc_str = line.split('=')[1]
+            loc = int(loc_str.strip())
+            found_loc = True
+
+    pars = {'order_wl': order_wl,
+            'ref_type': ref_type,
+            'loc': loc}
+    found_all = found_order & found_loc & found_ref
+
+    return pars, found_all
+
+
+def air2vac(air, unit='AA'):
+    # From Donald Morton 1991, ApJS 77,119
+    wl = air*u.Unit(unit)
+    wlum = wl.to(u.um).value
+    sigma2 = (1./wlum)**2
+    fact = 1.0 + 6.4328e-5 + 2.94981e-2/(146.0 - sigma2) + 2.5540e-4/(41.0 - sigma2)
+    vac = air*fact
+    return vac
+
+
+def vac2air(vac, unit='AA'):
+    # From Donald Morton 1991, ApJS 77,119
+    wl = vac*u.Unit(unit)
+    wlum = wl.to(u.um).value
+    sigma2 = (1./wlum)**2
+    fact = 1.0 + 6.4328e-5 + 2.94981e-2/(146.0 - sigma2) + 2.5540e-4/(41.0 - sigma2)
+    air = vac/fact
+    return air
+
+
+# def air2vac(wavelength, unit='AA'):
+#     """
+#     Implements the air to vacuum wavelength conversion described in eqn 65 of
+#     Griesen 2006
+#
+#     wavelength : array or float
+#         Input wavelength in air.
+#
+#     unit : string   [default='AA']
+#         Units of the input wavelengths, default is Angstrom (AA).
+#
+#     Returns
+#     -------
+#     The vacuum converted array of wavelength in the same units
+#     """
+#     wl = wavelength*u.Unit(unit)
+#     wlum = wl.to(u.um).value
+#     n_a = 1. + 1e-6*(287.6155 + 1.62887/wlum**2 + 0.01360/wlum**4)
+#     return n_a * wavelength
+#
+# def vac2air(wavelength, unit='AA'):
+#     """
+#     Implements the air to vacuum wavelength conversion described in eqn 65 of
+#     Griesen 2006
+#
+#     wavelength : array or float
+#         Input wavelength in vacuum.
+#
+#     unit : string   [default='AA']
+#         Units of the input wavelengths, default is Angstrom (AA).
+#
+#     Returns
+#     -------
+#     The air converted array of wavelength in the same units
+#     """
+#     wl = wavelength*u.Unit(unit)
+#     wlum = wl.to(u.um).value
+#     n_a = 1. + 1e-6*(287.6155 + 1.62887/wlum**2 + 0.01360/wlum**4)
+#     return wavelength / n_a
