@@ -1,5 +1,6 @@
 """
-  PyNOT Data Reduction Pipeline for NOT/ALFOSC
+  PyNOT Data Reduction Pipeline
+  Originally written for NOT/ALFOSC
 
 For available tasks, run:
     %] pynot -h
@@ -70,16 +71,36 @@ def set_help_width(max_width=30):
         return ArgumentDefaultsHelpFormatter
 
 
-def initialize(args):
+def initialize(path, mode, pfc_fname='dataset.pfc', pars_fname='options.yml', verbose=True):
+    """
+    Initialize new dataset with file classification table and default options.
+
+    path : str
+        Path to the folder containing the raw files to classify
+
+    mode : str  ['spex'/'spec' or 'phot']
+        The mode for the default options: `spec` for spectroscopy, and `phot` for photometry
+        `spex` is allowed for backwards compatibility.
+
+    pfc_fname : str
+        The filename of the file classification table
+
+    pars_filename : str
+        The filename of the default options
+
+    verbose : bool  [default=True]
+        Print logging to terminal
+    """
     print_credits()
     from pynot.data import organizer as do
     from pynot.data import io
 
     # Classify files:
     # args.silent is set to "store_false", so by default it is true!
-    verbose = args.silent
-    database, message = do.classify(args.path, progress=verbose)
-    pfc_fname = args.output
+    database, message = do.classify(path, progress=verbose)
+    if database is None:
+        print(message)
+        return
 
     if pfc_fname[-4:] != '.pfc':
         pfc_fname += '.pfc'
@@ -88,7 +109,7 @@ def initialize(args):
         print(message)
     print(" [OUTPUT] - Saved file classification database: %s" % pfc_fname)
 
-    if args.mode == 'spex':
+    if mode.lower() in ['spex', 'spec']:
         defaults_fname = defaults_fname_spec
     else:
         defaults_fname = defaults_fname_phot
@@ -110,9 +131,8 @@ def initialize(args):
     dataset_input = "%s:  %s #%s" % (root, fmt % pfc_fname, comment)
     all_lines[num] = dataset_input
 
-    pars_fname = args.pars
     if pars_fname == 'options.yml':
-        pars_fname = 'options_spex.yml' if args.mode == 'spex' else 'options_phot.yml'
+        pars_fname = 'options_spex.yml' if mode == 'spex' else 'options_phot.yml'
 
     if pars_fname[-4:] != '.yml':
         pars_fname += '.yml'
@@ -123,10 +143,11 @@ def initialize(args):
 
     print(" [OUTPUT] - Initiated new parameter file: %s\n" % pars_fname)
     print("\n You can now start the reduction pipeline by running:")
-    print("   ]%% pynot %s  %s\n" % (args.mode, pars_fname))
+    print("   ]%% pynot %s  %s\n" % (mode, pars_fname))
 
 
 def main():
+
     parser = ArgumentParser(prog='pynot')
     tasks = parser.add_subparsers(dest='task')
 
@@ -134,10 +155,10 @@ def main():
 
     parser_init = tasks.add_parser('init', formatter_class=set_help_width(31),
                                    help="Initiate a default parameter file")
-    parser_init.add_argument("mode", type=str, choices=['spex', 'phot'],
+    parser_init.add_argument("mode", type=str, choices=['spec', 'spex', 'phot'],
                              help="Create parameter file for spectroscopy or imaging?")
     parser_init.add_argument("path", type=str, nargs='+',
-                             help="Path (or paths) to raw ALFOSC data to be classified")
+                             help="Path (or paths) to raw data to be classified")
     parser_init.add_argument('-p', "--pars", type=str, default='options.yml',
                              help="Filename of parameter file, default = options.yml")
     parser_init.add_argument("-o", "--output", type=str, default='dataset.pfc',
@@ -145,7 +166,19 @@ def main():
     parser_init.add_argument("-s", "--silent", action='store_false',
                              help="Minimze the output to terminal")
 
+    parser_org = tasks.add_parser('classify', formatter_class=set_help_width(31),
+                                  help="Classify the files in `path`")
+    parser_org.add_argument("path", type=str, nargs='+',
+                            help="Path (or paths) to raw data to be classified")
+    parser_org.add_argument("-o", "--output", type=str, default='dataset.pfc',
+                            help="Filename of file classification table (*.pfc)")
+    parser_org.add_argument("-f", "--force", action="store_true",
+                            help="Force overwrite of the file classification table (*.pfc)")
 
+    parser_obd = tasks.add_parser('update-obs', formatter_class=set_help_width(31),
+                                  help="Update the OB database based on a classification table")
+    parser_obd.add_argument("pfc", type=str,
+                            help="File classification table (*.pfc)")
 
     # -- BIAS :: Bias Combination
     parser_bias = tasks.add_parser('bias', formatter_class=set_help_width(31),
@@ -166,8 +199,8 @@ def main():
                               help="Input file containing list of image filenames to combine")
     parser_sflat.add_argument("--bias", type=str, required=True,
                               help="Filename of combined bias frame  [REQUIRED]")
-    parser_sflat.add_argument("-o", "--output", type=str, required=True,
-                              help="Output filename of combined bias frame  [REQUIRED]")
+    parser_sflat.add_argument("-o", "--output", type=str, default='',
+                              help="Output filename of combined and normalized flat frame. Constructed from GRISM and SLIT by default")
     parser_sflat.add_argument("--axis", type=int, default=2,
                               help="Dispersion axis: 1 horizontal, 2: vertical")
     # Define parameters based on default values:
@@ -195,10 +228,10 @@ def main():
                              help="List of filenames to correct")
     parser_corr.add_argument("--dir", type=str, default='',
                              help="Output directory")
-    parser_corr.add_argument("--bias", type=str, required=True,
-                             help="Filename of combined bias frame  [REQUIRED]")
+    parser_corr.add_argument("--bias", type=str, default='',
+                             help="Filename of combined bias frame. If not given, a constant of 0 is used")
     parser_corr.add_argument("--flat", type=str, default='',
-                             help="Filename of combined flat frame")
+                             help="Filename of combined flat frame. If not given, a constant of 1 is used")
     parser_corr.add_argument("--img", action='store_true',
                              help="Imaging mode")
 
@@ -206,7 +239,7 @@ def main():
     # -- identify :: Identify Arc Lines
     parser_id = tasks.add_parser('identify', formatter_class=set_help_width(31),
                                  help="Interactive identification of arc lines")
-    parser_id.add_argument("arc", type=str,
+    parser_id.add_argument("arc", type=str, default='', nargs='?',
                            help="Input filename of arc line image")
     parser_id.add_argument("--lines", type=str, default='',
                            help="Linelist, automatically loaded if possible")
@@ -214,12 +247,16 @@ def main():
                            help="Dispersion axis: 1 horizontal, 2: vertical")
     parser_id.add_argument("-o", "--output", type=str, default='',
                            help="Output filename of arc line identification table")
+    parser_id.add_argument("--air", action='store_true',
+                           help="Use air reference wavelengths")
+    parser_id.add_argument("--loc", type=int, default=-1,
+                           help="Location along the slit to extract lamp spectrum [pixels].")
 
 
     # -- response :: Calculate Response Function
     parser_resp = tasks.add_parser('response', formatter_class=set_help_width(31),
                                    help="Interactive determination of instrument response function")
-    parser_resp.add_argument("input", type=str,
+    parser_resp.add_argument("input", type=str, default='', nargs='?',
                              help="Input filename of 1D spectrum of flux standard star")
     parser_resp.add_argument("-o", "--output", type=str, default='',
                              help="Output filename of response function")
@@ -286,9 +323,9 @@ def main():
     parser_crr.add_argument('-n', "--niter", type=int, default=4,
                             help="Number of iterations")
     parser_crr.add_argument("--gain", type=float, default=0.16,
-                            help="Detector gain, default for ALFOSC CCD14: 0.16 e-/ADU")
+                            help="Detector gain  (e-/ADU)")
     parser_crr.add_argument("--readnoise", type=float, default=4.3,
-                            help="Detector read noise, default for ALFOSC CCD14: 4.3 e-")
+                            help="Detector read noise (e-)")
     # Define parameters based on default values:
     set_default_pars(parser_crr, section='crr', default_type=int,
                      ignore_pars=['niter', 'gain', 'readnoise'])
@@ -316,10 +353,28 @@ def main():
                               help="Output filename of flux-calibrated 2D spectrum [REQUIRED]")
 
 
+    # -- scombine :: Spectral combination in 1D or 2D
+    parser_scomb = tasks.add_parser('scombine', formatter_class=set_help_width(31),
+                                    help="Spectral combination")
+    parser_scomb.add_argument("input", type=str, nargs='+',
+                              help="Input spectra")
+    parser_scomb.add_argument("-o", "--output", type=str, default='',
+                              help="Output filename (autogenerated if not given)")
+    parser_scomb.add_argument("-m", "--method", type=str,
+                              help="Combination method", default='mean',
+                              choices=['mean', 'median'])
+    parser_scomb.add_argument("-s", "--scale", action="store_true",
+                              help="Scale spectra before combining")
+    parser_scomb.add_argument("--axis", choices=[1, 2], default=1,
+                              help="Dispersion axis. 1: horizontal, 2: vertical (default: 1 for processed spectra)")
+    parser_scomb.add_argument("-x", "--extended", action="store_true",
+                              help="Set this option for 2D spectra of extended sources to turn off automatic localization")
+
+
     # -- extract :: Extraction of 1D spectrum from 2D
     parser_ext = tasks.add_parser('extract', formatter_class=set_help_width(31),
                                   help="Extract 1D spectrum from 2D")
-    parser_ext.add_argument("input", type=str,
+    parser_ext.add_argument("input", type=str, default='', nargs='?',
                             help="Input filename of 2D spectrum")
     parser_ext.add_argument("-o", "--output", type=str, default='',
                             help="Output filename of 1D spectrum (FITS Table)")
@@ -331,6 +386,7 @@ def main():
     set_default_pars(parser_ext, section='extract', default_type=int,
                      ignore_pars=['interactive'])
 
+
     # Spectral Redux:
     parser_redux = tasks.add_parser('spex', formatter_class=set_help_width(30),
                                     help="Run the full spectroscopic pipeline")
@@ -342,6 +398,14 @@ def main():
                               help="Minimze the output to terminal")
     parser_redux.add_argument("-i", "--interactive", action="store_true",
                               help="Use interactive interface throughout")
+    parser_redux.add_argument("-f", "--force", action="store_true",
+                              help="Force restart of all OBs!")
+    parser_redux.add_argument("--mbias", action="store_true",
+                              help="Run bias only")
+    parser_redux.add_argument("--mflat", action="store_true",
+                              help="Run flat combination and normalization only")
+    parser_redux.add_argument("--arcs", action="store_true",
+                              help="Process arcs")
 
     parser_break = tasks.add_parser('', help="")
 
@@ -352,6 +416,8 @@ def main():
                              help="Input filename of pipeline configuration in YAML format")
     parser_phot.add_argument("-s", "--silent", action="store_false",
                              help="Minimze the output to terminal")
+    parser_phot.add_argument("-f", "--force", action="store_true",
+                             help="Force restart of all OBs!")
 
     parser_imtrim = tasks.add_parser('imtrim', formatter_class=set_help_width(30),
                                      help="Trim images")
@@ -438,6 +504,34 @@ def main():
     parser_findnew.add_argument('-z', "--zp", type=float,
                                 help="Magnitude zero point in case the source catalog has not been flux calibrated")
 
+    parser_setup = tasks.add_parser('setup', formatter_class=set_help_width(30),
+                                    help="Install new instrument configuration")
+    parser_setup.add_argument('module', type=str,
+                              help='Filename of the instrument configuration module (*.py)')
+    parser_setup.add_argument('-f', '--filters', type=str, default='',
+                              help='Filename of the instrument filter definitions (must have columns: name and short_name)')
+    parser_setup.add_argument('-r', '--rules', type=str, default='',
+                              help='Filename of the classification ruleset')
+    parser_setup.add_argument('--ext', type=str, default='',
+                              help='Filename of the observatory extinction data (lapalma, lasilla, paranal, or file path)')
+    parser_setup.add_argument('-u', '--use', action="store_true",
+                              help='Switch immediately to use the newly installed instrument')
+
+    parser_use = tasks.add_parser('use', formatter_class=set_help_width(30),
+                                  help="Set current instrument")
+    parser_use.add_argument('name', type=str, default='', nargs='?',
+                            help='Name of the instrument to use (see --list)')
+    parser_use.add_argument('--list', action="store_true",
+                            help='Display a list of the currently installed instruments')
+
+    parser_f2a = tasks.add_parser('fits2ascii', formatter_class=set_help_width(30),
+                                  help="Convert FITS spectrum to ASCII table format")
+    parser_f2a.add_argument('input', type=str,
+                            help='Input filename of FITS spectrum (either table or MEF)')
+    parser_f2a.add_argument('output', type=str,
+                            help='Output filename of the ASCII table')
+    parser_f2a.add_argument('--keys', type=str, nargs='+',
+                            help='List of keywords from the FITS header to include (ESO keywords must use . not space delimiter)')
 
     args = parser.parse_args()
 
@@ -448,7 +542,11 @@ def main():
 
 
     if task == 'init':
-        initialize(args)
+        initialize(args.path, args.mode, pfc_fname=args.output, pars_fname=args.pars, verbose=args.silent)
+
+    elif task == 'update-obs':
+        from pynot.data.obs import update_ob_database
+        update_ob_database(args.pfc)
 
     elif task == 'spex':
         from pynot.redux import run_pipeline
@@ -456,7 +554,11 @@ def main():
         run_pipeline(options_fname=args.params,
                      object_id=args.object,
                      verbose=args.silent,
-                     interactive=args.interactive)
+                     interactive=args.interactive,
+                     force_restart=args.force,
+                     make_bias=args.mbias,
+                     make_flat=args.mflat,
+                     make_arcs=args.arcs)
 
     elif task == 'bias':
         from pynot.calibs import combine_bias_frames
@@ -476,6 +578,8 @@ def main():
         for varname in vars_to_remove:
             options.pop(varname)
         _, log = normalize_spectral_flat(flatcombine, args.output, dispaxis=args.axis, **options)
+        # from pynot.calibs import task_sflat
+        # task_sflat(args)
 
     elif task == 'corr':
         from pynot.scired import correct_raw_file
@@ -490,6 +594,7 @@ def main():
         else:
             input_list = np.loadtxt(args.input, dtype=str, usecols=(0,))
 
+        # Mode determines the header keywords to update (CDELT or CD-matrix)
         if args.img:
             mode = 'img'
         else:
@@ -503,7 +608,7 @@ def main():
             if args.dir != '':
                 output = os.path.join(args.dir, output)
             _ = correct_raw_file(fname, output=output, bias_fname=args.bias, flat_fname=args.flat,
-                                 overscan=50, overwrite=True, mode=mode)
+                                 overwrite=True, mode=mode)
             print(" - Image: %s  ->  %s" % (fname, output))
 
     elif task == 'identify':
@@ -514,6 +619,8 @@ def main():
         gui = GraphicInterface(args.arc,
                                linelist_fname=args.lines,
                                dispaxis=args.axis,
+                               air=args.air,
+                               loc=args.loc,
                                output=args.output)
         gui.show()
         app.exit(app.exec_())
@@ -547,7 +654,7 @@ def main():
         from pynot.scired import auto_fit_background
         print("Running task: Background Subtraction")
         options = copy(vars(args))
-        vars_to_remove = ['task', 'input', 'output', 'axis']
+        vars_to_remove = ['task', 'input', 'output', 'axis', 'auto']
         for varname in vars_to_remove:
             options.pop(varname)
         log = auto_fit_background(args.input, args.output, dispaxis=args.axis,
@@ -566,12 +673,51 @@ def main():
     elif task == 'flux1d':
         from pynot.response import flux_calibrate_1d
         print("Running task: Flux Calibration of 1D Spectrum")
-        log = flux_calibrate_1d(args.input, output=args.output, response=args.response)
+        log = flux_calibrate_1d(args.input, output=args.output, response_fname=args.response)
 
     elif task == 'flux2d':
         from pynot.response import flux_calibrate
         print("Running task: Flux Calibration of 2D Image")
-        log = flux_calibrate(args.input, output=args.output, response=args.response)
+        log = flux_calibrate(args.input, output=args.output, response_fname=args.response)
+
+    elif task == 'scombine':
+        print("Running task: Spectral Combination")
+        from pynot.scombine import combine_1d, combine_2d
+        from astropy.io import fits
+        from glob import glob
+
+        if len(args.input) == 1:
+            input_arg = args.input[0]
+            if '*' in input_arg or '?' in input_arg:
+                filelist = glob.glob(input_arg)
+            else:
+                filelist = np.loadtxt(input_arg, usecols=(0,), dtype=str)
+        else:
+            filelist = args.input
+
+        # Check if data are 1D or 2D:
+        try:
+            data = fits.getdata(filelist[0])
+            if isinstance(data, fits.fitsrec.FITS_rec):
+                data_is_1d = True
+            elif isinstance(data, np.ndarray) and len(data.shape) == 1:
+                data_is_1d = True
+            elif isinstance(data, np.ndarray) and len(data.shape) == 2:
+                data_is_1d = False
+            else:
+                print("  [ERROR] - Could not recognize the data type!")
+
+        except OSError:
+            data_is_1d = True
+
+
+        if data_is_1d:
+            out_args = combine_1d(filelist, output=args.output, method=args.method, scale=args.scale)
+        else:
+            out_args = combine_2d(filelist, output=args.output, method=args.method,
+                                  scale=args.scale, extended=args.extended, dispaxis=args.axis)
+        log = out_args[-1]
+
 
     elif task == 'extract':
         from PyQt5 import QtWidgets
@@ -593,10 +739,12 @@ def main():
             app.exit(app.exec_())
 
 
+    # -- Imaging tasks:
     elif task == 'phot':
         from pynot.phot_redux import run_pipeline
         run_pipeline(options_fname=args.params,
-                     verbose=args.silent)
+                     verbose=args.silent,
+                     force_restart=args.force)
 
     elif task == 'imflat':
         print("Running task: Combination of Imaging Flat Fields")
@@ -695,6 +843,63 @@ def main():
                                             mag_lim=args.limit,
                                             zp=args.zp)
 
+    elif task == 'setup':
+        print("\nInstalling new instrument settings...")
+        from pynot.insconfig import setup_instrument
+        log = setup_instrument(args)
+
+    elif task == 'use':
+        from pynot.insconfig import get_installed_instruments, change_instrument
+
+        all_instruments, inst_list = get_installed_instruments()
+        if args.list:
+            print(inst_list)
+            return
+        else:
+            log = change_instrument(args.name, all_instruments)
+
+    elif task == 'classify':
+        print_credits()
+        from pynot.data import organizer as do
+        from pynot.data import io
+
+        # Classify files:
+        database, message = do.classify(args.path)
+        print("")
+        print(message)
+        if database is None:
+            return
+
+        if not args.output.endswith('.pfc'):
+            pfc_fname = args.output + '.pfc'
+        else:
+            pfc_fname = args.output
+
+        if os.path.exists(pfc_fname):
+            print("          - File classification database already exists")
+            if args.force:
+                print("          - Overwriting the database! Are you sure?  [Y/n]")
+                user_input = input("          > ")
+                if user_input.lower() in ['y', 'yes', '']:
+                    pass
+                else:
+                    print("Aborting...")
+                    return
+            else:
+                previous_database = io.load_database(pfc_fname)
+                database = previous_database + database
+                print("          - Merging the databases")
+        io.save_database(database, pfc_fname)
+        print(" [OUTPUT] - Saved file classification database: %s" % pfc_fname)
+
+    elif task == 'fits2ascii':
+        from pynot.fitsio import fits_to_ascii
+        fits_to_ascii(args.input, args.output, args.keys)
+
+    else:
+        import pynot
+        print("Running PyNOT for instrument: %s\n" % pynot.instrument.name)
+
     if log:
         print(log)
 
