@@ -6,17 +6,18 @@ from astropy.io import fits
 from collections import defaultdict
 import os
 import sys
-import datetime
 import numpy as np
 
-from pynot import alfosc
+from pynot import instrument
 from pynot.data import io
 from pynot.data import organizer as do
+from pynot.data import obs
 from pynot.phot import image_combine, create_fringe_image, source_detection, flux_calibration_sdss
 from pynot.calibs import combine_bias_frames, combine_flat_frames
 from pynot.functions import get_options, get_version_number
 from pynot.scired import raw_correction, correct_cosmics, trim_filter_edge, detect_filter_edge
 from pynot.wcs import correct_wcs
+from pynot.logging import Report
 
 from PyQt5.QtWidgets import QApplication
 
@@ -26,96 +27,8 @@ defaults_fname = os.path.join(calib_dir, 'default_options_img.yml')
 __version__ = get_version_number()
 
 
-class Report(object):
-    def __init__(self, verbose=False):
-        self.verbose = verbose
-        self.time = datetime.datetime.now()
-        self.fname = 'pynot_img_%s.log' % self.time.strftime('%d%b%Y-%Hh%Mm%S')
-        self.remarks = list()
-        self.lines = list()
-        self.header = """
-         PyNOT Data Processing Pipeline
-        ================================
-        version %s
-        %s
 
-        """ % (__version__, self.time.strftime("%b %d, %Y  %H:%M:%S"))
-        self.report = ""
-
-        if self.verbose:
-            print(self.header)
-
-    def clear(self):
-        self.lines = list()
-        self.remarks = list()
-
-    def set_filename(self, fname):
-        self.fname = fname
-
-    def commit(self, text):
-        if self.verbose:
-            print(text, end='', flush=True)
-        self.lines.append(text)
-
-    def error(self, text):
-        text = ' [ERROR]  - ' + text
-        if self.verbose:
-            print(text)
-        if text[-1] != '\n':
-            text += '\n'
-        self.lines.append(text)
-
-    def warn(self, text, force=False):
-        text = '[WARNING] - ' + text
-        if self.verbose or force:
-            print(text)
-        if text[-1] != '\n':
-            text += '\n'
-        self.lines.append(text)
-
-    def write(self, text, prefix='          - '):
-        text = prefix + text
-        if self.verbose:
-            print(text)
-        if text[-1] != '\n':
-            text += '\n'
-        self.lines.append(text)
-
-    def add_linebreak(self):
-        if self.verbose:
-            print("")
-        self.lines.append("\n")
-
-    def add_remark(self, text):
-        self.remarks.append(text)
-
-    def _make_report(self):
-        remark_str = ''.join(self.remarks)
-        lines_str = ''.join(self.lines)
-        self.report = '\n'.join([self.header, remark_str, lines_str])
-
-    def print_report(self):
-        self._make_report()
-        print(self.report)
-
-    def save(self):
-        self._make_report()
-        with open(self.fname, 'w') as output:
-            output.write(self.report)
-
-    def exit(self):
-        print(" - Pipeline terminated.")
-        print(" Consult the log: %s\n" % self.fname)
-        self.save()
-
-    def fatal_error(self):
-        print(" !! FATAL ERROR !!")
-        print(" Consult the log: %s\n" % self.fname)
-        self.save()
-
-
-
-def run_pipeline(options_fname, verbose=False):
+def run_pipeline(options_fname, verbose=False, force_restart=False):
     log = Report(verbose)
 
     global app
@@ -151,12 +64,27 @@ def run_pipeline(options_fname, verbose=False):
         log.fatal_error()
         return
     object_filelist = database['IMG_OBJECT']
-    raw_image_list = list(map(do.RawImage, object_filelist))
+    try:
+        raw_image_list = list(map(do.RawImage, object_filelist))
+    except (ValueError, do.UnknownObservingMode, OSError, FileNotFoundError, TypeError, IndexError) as e:
+        log.error(str(e))
+        log.fatal_error()
+        return
 
     object_images = defaultdict(lambda: defaultdict(list))
     for sci_img in raw_image_list:
         object_images[sci_img.target_name][sci_img.filter].append(sci_img)
 
+    obd_fname = os.path.splitext(dataset_fname)[0] + '.obd'
+    obdb = obs.OBDatabase(obd_fname)
+    if os.path.exists(obd_fname):
+        log.write("Loaded OB database: %s" % obd_fname)
+    else:
+        log.write("Initiated OB database: %s" % obd_fname)
+    obdb.update_imaging(object_images)
+    log.write("Updating OB database")
+    log.add_linebreak()
+
     # get list of unique filters in dataset:
     filter_list = np.unique(sum([list(obj.keys()) for obj in object_images.values()], []))
 
@@ -170,12 +98,12 @@ def run_pipeline(options_fname, verbose=False):
 
     flat_images = database['IMG_FLAT']
     for flat_file in flat_images:
-        this_filter = do.get_filter(fits.getheader(flat_file))
+        this_filter = instrument.get_filter(fits.getheader(flat_file))
         if this_filter in filter_list:
             flat_images_for_filter[this_filter].append(flat_file)
 
     # All files are put in a folder: imaging/OBJNAME/filter/...
-    output_base = 'imaging'
+    output_base = obs.output_base_phot
     if not os.path.exists(output_base):
         os.mkdir(output_base)
 
@@ -208,8 +136,7 @@ def run_pipeline(options_fname, verbose=False):
     master_flat = {}
     filter_edges = {}
     log.write("Running task: Imaging Flat Combination")
-    for filter_raw, flat_frames in flat_images_for_filter.items():
-        filter_name = alfosc.filter_translate[filter_raw]
+    for filter_name, flat_frames in flat_images_for_filter.items():
         log.write("Combining images for filter: %s" % filter_name)
         comb_flat_fname = os.path.join(output_base, 'FLAT_%s.fits' % filter_name)
         try:
@@ -218,7 +145,7 @@ def run_pipeline(options_fname, verbose=False):
                                               method=options['flat']['method'],
                                               overwrite=True, mode='img')
             log.commit(flat_msg)
-            master_flat[filter_raw] = comb_flat_fname
+            master_flat[filter_name] = comb_flat_fname
         except:
             log.error("Flat field combination failed for filter: %s" % filter_name)
             log.fatal_error()
@@ -231,7 +158,7 @@ def run_pipeline(options_fname, verbose=False):
             log.write("Detected edges on X-axis: %i  ;  %i" % (x1, x2))
             log.write("Detected edges on Y-axis: %i  ;  %i" % (y1, y2))
             log.add_linebreak()
-            filter_edges[filter_raw] = (x1, x2, y1, y2)
+            filter_edges[filter_name] = (x1, x2, y1, y2)
         except:
             log.error("Automatic edge detection failed!")
             log.fatal_error()
@@ -261,10 +188,15 @@ def run_pipeline(options_fname, verbose=False):
 
         log.write("Target Name: %s" % target_name, prefix=' [TARGET] - ')
 
-        for filter_raw, image_list in images_per_filter.items():
+        for filter_name, image_list in images_per_filter.items():
             # Create working directory:
-            filter_name = alfosc.filter_translate[filter_raw]
             output_dir = os.path.join(output_obj_base, filter_name)
+            if obdb.data[output_dir] in ['DONE', 'SKIP'] and not force_restart:
+                log.write("Skipping OB: %s  (status=%s)" % (output_dir, obdb.data[output_dir]))
+                log.write("Change OB status to blank in the .obd file if you want to redo the reduction")
+                log.write("or run the pipeline with the '-f' option to force re-reduction of all OBs")
+                log.add_linebreak()
+                continue
             if not os.path.exists(output_dir):
                 os.mkdir(output_dir)
             log.write("Filter : %s" % filter_name)
@@ -285,7 +217,7 @@ def run_pipeline(options_fname, verbose=False):
                 flat_fname = master_flat[sci_img.filter]
                 try:
                     _ = raw_correction(sci_img.data, sci_img.header, master_bias_fname, flat_fname,
-                                       output=corrected_fname, overwrite=True, overscan=50, mode='img')
+                                       output=corrected_fname, overwrite=True, mode='img')
                     log.commit("          - bias+flat ")
                     temp_images.append(corrected_fname)
                 except:
@@ -295,7 +227,7 @@ def run_pipeline(options_fname, verbose=False):
                     raise
 
                 # Trim edges:
-                image_region = filter_edges[filter_raw]
+                image_region = filter_edges[filter_name]
                 try:
                     _ = trim_filter_edge(corrected_fname, *image_region, output=trim_fname)
                     log.commit(" trim ")
@@ -329,8 +261,8 @@ def run_pipeline(options_fname, verbose=False):
             # Create Fringe image:
             if options['skysub']['defringe'] and N_images > 3:
                 log.write("Running task: Creating Average Fringe Image")
-                fringe_fname = os.path.join(output_dir, 'fringe_image.fits')
-                fringe_pdf_fname = os.path.join(output_dir, 'fringe_image.pdf')
+                fringe_fname = os.path.join(output_dir, 'fringe_image_%s.fits' % filter_name)
+                fringe_pdf_fname = os.path.join(output_dir, 'fringe_image_%s.pdf' % filter_name)
                 try:
                     msg = create_fringe_image(corrected_images, output=fringe_fname, fig_fname=fringe_pdf_fname,
                                               threshold=3)
@@ -341,14 +273,16 @@ def run_pipeline(options_fname, verbose=False):
                     log.fatal_error()
                     print("Unexpected error:", sys.exc_info()[0])
                     raise
+            elif options['skysub']['defringe'] and N_images <= 3:
+                log.warn("No fringe image can be created. Need at least 3 images.")
             else:
                 fringe_fname = ''
 
 
             # Combine individual images for a given filter:
             if len(image_list) > 50:
-                log.warning("Large amounts of memory needed for image combination!", force=True)
-                log.warning("A total of %i images will be combined." % len(image_list), force=True)
+                log.warn("Large amounts of memory needed for image combination!", force=True)
+                log.warn("A total of %i images will be combined." % len(image_list), force=True)
 
             log.write("Running task: Image Combination")
             comb_log_name = os.path.join(output_dir, 'filelist_%s.txt' % target_name)
@@ -358,9 +292,13 @@ def run_pipeline(options_fname, verbose=False):
                                            fringe_image=fringe_fname, **options['combine'])
                 log.commit(output_msg)
                 log.add_linebreak()
-            except:
+            except (IndexError, FileNotFoundError, OSError) as e:
                 log.error("Image combination failed!")
+                log.error(str(e))
                 log.fatal_error()
+                return
+            except:
+                log.datal_error()
                 print("Unexpected error:", sys.exc_info()[0])
                 raise
 
@@ -414,5 +352,6 @@ def run_pipeline(options_fname, verbose=False):
                     log.write(fname)
                 log.add_linebreak()
 
+            obdb.update(output_dir, 'DONE')
 
     log.exit()
