@@ -15,23 +15,24 @@ import os
 import re
 import sys
 import numpy as np
-import matplotlib
 from matplotlib.backends.backend_qt5agg import FigureCanvas, NavigationToolbar2QT as NavigationToolbar
 from matplotlib.figure import Figure
 from scipy.interpolate import UnivariateSpline
-from numpy.polynomial import Chebyshev
 from astropy.io import fits
-from PyQt5 import QtCore, QtGui, QtWidgets
+from PyQt5 import QtGui, QtWidgets
 import warnings
 
-from pynot import alfosc
+from pynot import instrument
+from pynot import response
 from pynot.functions import get_version_number
+from pynot.welcome import WelcomeMessage
 
 
+code_dir = os.path.dirname(os.path.abspath(__file__))
 __version__ = get_version_number()
 
 
-def run_gui(input_fname, output_fname, app=None, order=3, smoothing=0.02):
+def run_gui(input_fname, output_fname='', app=None, order=3, smoothing=0.02):
     if app is None:
         app = QtWidgets.QApplication(sys.argv)
     gui = ResponseGUI(input_fname, output_fname=output_fname, locked=True, order=order, smoothing=smoothing)
@@ -61,6 +62,7 @@ class ResponseGUI(QtWidgets.QMainWindow):
         self.setWindowTitle('PyNOT: Response')
         self._main = QtWidgets.QWidget()
         self.setCentralWidget(self._main)
+        self.welcome_msg = None
 
         # Set attributes:
         self.spectrum = None
@@ -69,9 +71,9 @@ class ResponseGUI(QtWidgets.QMainWindow):
         self.output_fname = output_fname
         self.first_time_open = True
         # Extinction table attributes:
-        self.ext_fname = alfosc.path + '/calib/lapalma.ext'
+        self.ext_fname = instrument.extinction_fname
         try:
-            ext_wl, ext = np.loadtxt(alfosc.path + '/calib/lapalma.ext', unpack=True)
+            ext_wl, ext = np.loadtxt(self.ext_fname, unpack=True)
         except:
             ext_wl, ext = None, None
         self.ext_wl = ext_wl
@@ -87,9 +89,9 @@ class ResponseGUI(QtWidgets.QMainWindow):
 
         # Fitting Parameters:
         self.star_chooser = QtWidgets.QComboBox()
-        self.all_names = sorted([name.upper() for name in alfosc.standard_stars] + [''])
+        self.all_names = sorted([name.lower() for name in response.standard_stars] + [''])
         self.star_chooser.addItems(self.all_names)
-        self.star_chooser.setCurrentText(star_name)
+        self.star_chooser.setCurrentText('')
         self.star_chooser.currentTextChanged.connect(self.set_star)
 
         self.exptime_edit = QtWidgets.QLineEdit("")
@@ -101,7 +103,7 @@ class ResponseGUI(QtWidgets.QMainWindow):
         self.order_edit.setValidator(QtGui.QIntValidator(1, 5))
         self.order_edit.returnPressed.connect(self.fit_response)
 
-        self.smooth_edit = QtWidgets.QLineEdit("%.2f" % smoothing)
+        self.smooth_edit = QtWidgets.QLineEdit("%.3f" % smoothing)
         self.smooth_edit.setValidator(QtGui.QDoubleValidator())
         self.smooth_edit.returnPressed.connect(self.fit_response)
 
@@ -123,19 +125,14 @@ class ResponseGUI(QtWidgets.QMainWindow):
         self.response_points = None
 
         # == TOP MENU BAR:
-        self.save_btn = QtWidgets.QPushButton("Save")
-        self.save_btn.clicked.connect(self.save_response)
+        # self.save_btn = QtWidgets.QPushButton("Save")
+        # self.save_btn.clicked.connect(self.save_response)
         self.load_btn = QtWidgets.QPushButton("Load")
         self.load_btn.clicked.connect(self.load_spectrum)
+        self.close_btn = QtWidgets.QPushButton("Done")
+        self.close_btn.clicked.connect(lambda x: self.done(locked=locked))
         if locked:
-            self.close_btn = QtWidgets.QPushButton("Done")
-            self.close_btn.clicked.connect(self.done)
             self.load_btn.setEnabled(False)
-            self.save_btn.setEnabled(False)
-            self.save_btn.setText("")
-        else:
-            self.close_btn = QtWidgets.QPushButton("Close")
-            self.close_btn.clicked.connect(self.close)
 
 
         # == Layout ===========================================================
@@ -145,7 +142,7 @@ class ResponseGUI(QtWidgets.QMainWindow):
 
         top_menubar = QtWidgets.QHBoxLayout()
         top_menubar.addWidget(self.close_btn)
-        top_menubar.addWidget(self.save_btn)
+        # top_menubar.addWidget(self.save_btn)
         top_menubar.addWidget(self.load_btn)
         top_menubar.addStretch(1)
 
@@ -228,15 +225,67 @@ class ResponseGUI(QtWidgets.QMainWindow):
 
         self.create_menu()
 
+        self.show()
         # -- Set Data:
         if fname:
             self.load_spectrum(fname)
+            self.show_welcome(has_file=True)
+        else:
+            self.show_welcome(has_file=False)
+
+        if star_name:
+            self.star_name = star_name
+            self.star_chooser.setCurrentText(star_name)
 
 
-    def done(self):
-        success = self.save_response(self.output_fname)
-        if success:
-            self.close()
+    def show_welcome(self, has_file=False, force=False):
+        if self.welcome_msg is not None:
+            self.welcome_msg.close()
+
+        cache_file = os.path.join(code_dir, '.response_msg')
+        if os.path.exists(cache_file):
+            if not force:
+                return
+
+        html_file = code_dir + '/data/help/welcome_msg_response.html'
+        self.welcome_msg = WelcomeMessage(cache_file, html_file,
+                                          has_file=has_file)
+        self.welcome_msg.exec_()
+
+
+    def done(self, locked=False):
+        msg = "You're about to quit PyNOT: Response"
+        messageBox = QtWidgets.QMessageBox()
+        messageBox.setText(msg)
+        if self.welcome_msg is not None:
+            self.welcome_msg.close()
+
+        if locked:
+            info_msg = "Save the fitted response function and move on?"
+            info_msg += "\nCurrent filename: %s" % self.output_fname
+            messageBox.setStandardButtons(QtWidgets.QMessageBox.Cancel | QtWidgets.QMessageBox.Save)
+            messageBox.setInformativeText(info_msg)
+            messageBox.setDefaultButton(QtWidgets.QMessageBox.Save)
+            retval = messageBox.exec_()
+            if retval == QtWidgets.QMessageBox.Save:
+                success = self.save_response(self.output_fname)
+                if success:
+                    self.close()
+
+        else:
+            info_msg = "Do you want to save the fitted response function before quitting?"
+            messageBox.setStandardButtons(QtWidgets.QMessageBox.Discard | QtWidgets.QMessageBox.Cancel | QtWidgets.QMessageBox.Save)
+            messageBox.setInformativeText(info_msg)
+            messageBox.setDefaultButton(QtWidgets.QMessageBox.Save)
+            retval = messageBox.exec_()
+            if retval == QtWidgets.QMessageBox.Save:
+                success = self.save_response(self.output_fname)
+                if success:
+                    self.close()
+
+            elif retval == QtWidgets.QMessageBox.Discard:
+                self.close()
+
 
     def save_response(self, fname=''):
         if self.response is None:
@@ -245,7 +294,8 @@ class ResponseGUI(QtWidgets.QMainWindow):
             return False
 
         if not fname:
-            current_dir = os.path.dirname(os.path.abspath(__file__))
+            # current_dir = os.path.dirname(os.path.abspath(__file__))
+            current_dir = './'
             basename = os.path.join(current_dir, "response_%s.fits" % (self.spectrum.header['OBJECT']))
             filters = "FITS Files (*.fits *.fit)"
             fname, _ = QtWidgets.QFileDialog.getSaveFileName(self, 'Save Response Function', basename, filters)
@@ -254,14 +304,16 @@ class ResponseGUI(QtWidgets.QMainWindow):
             hdu = fits.HDUList()
             prim_hdr = fits.Header()
             prim_hdr['AUTHOR'] = 'PyNOT version %s' % __version__
-            prim_hdr['OBJECT'] = self.spectrum.header['OBJECT']
-            prim_hdr['DATE-OBS'] = self.spectrum.header['DATE-OBS']
-            prim_hdr['EXPTIME'] = self.spectrum.header['EXPTIME']
-            prim_hdr['AIRMASS'] = self.spectrum.header['AIRMASS']
-            prim_hdr['ALGRNM'] = self.spectrum.header['ALGRNM']
-            prim_hdr['ALAPRTNM'] = self.spectrum.header['ALAPRTNM']
+            prim_hdr['INSTRUME'] = 'PyNOT'
+            prim_hdr['OBJECT'] = instrument.get_object(self.spectrum.header)
+            prim_hdr['DATE-OBS'] = instrument.get_date(self.spectrum.header)
+            prim_hdr['EXPTIME'] = float(self.exptime_edit.text())
+            prim_hdr['AIRMASS'] = float(self.airmass_edit.text())
+            prim_hdr['GRISM'] = instrument.get_grism(self.spectrum.header)
+            prim_hdr['SLIT'] = instrument.get_slit(self.spectrum.header)
             prim_hdr['RA'] = self.spectrum.header['RA']
             prim_hdr['DEC'] = self.spectrum.header['DEC']
+            prim_hdr['STD-STAR'] = self.star_name
             prim_hdr['COMMENT'] = 'PyNOT response function'
             prim = fits.PrimaryHDU(header=prim_hdr)
             hdu.append(prim)
@@ -286,16 +338,21 @@ class ResponseGUI(QtWidgets.QMainWindow):
         self.response = None
         self.mask = None
         self.filename = ""
+        self.data_line = None
+        self.fit_line = None
+        self.data_points = None
+        self.response_points = None
         self.canvas_points.draw()
 
 
     def load_spectrum(self, fname=''):
         if fname is False:
-            current_dir = os.path.dirname(os.path.abspath(__file__))
+            current_dir = './'
             filters = "FITS files (*.fits | *.fit)"
             fname, _ = QtWidgets.QFileDialog.getOpenFileName(self, 'Open 1D Spectrum', current_dir, filters)
             fname = str(fname)
             if self.first_time_open:
+                print("")
                 print(" [INFO] - Don't worry about the warning above. It's an OS warning that can not be suppressed.")
                 print("          Everything works as it should")
                 self.first_time_open = False
@@ -311,27 +368,25 @@ class ResponseGUI(QtWidgets.QMainWindow):
         self.spectrum = Spectrum(wl=table['WAVE'], data=table['FLUX'], header=hdr,
                                  wl_unit=table.columns['WAVE'].unit,
                                  flux_unit=table.columns['FLUX'].unit)
-        if 'EXPTIME' in hdr:
-            self.exptime_edit.setText("%.1f" % hdr['EXPTIME'])
+
+        if instrument.get_exptime(hdr):
+            self.exptime_edit.setText("%.1f" % instrument.get_exptime(hdr))
             self.exptime_edit.setEnabled(False)
         else:
             self.exptime_edit.setEnabled(True)
-        if 'AIRMASS' in hdr:
-            self.airmass_edit.setText("%.1f" % hdr['AIRMASS'])
+        if instrument.get_airmass(hdr):
+            self.airmass_edit.setText("%.3f" % instrument.get_airmass(hdr))
             self.airmass_edit.setEnabled(False)
         else:
             self.airmass_edit.setEnabled(True)
 
-        if 'TCSTGT' in hdr:
-            TCSname = hdr['TCSTGT']
-            TCSname = alfosc.lookup_std_star(TCSname)
-            if TCSname:
-                star_name = alfosc.standard_star_names[TCSname]
-                self.star_chooser.setCurrentText(star_name.upper())
-        elif 'OBJECT' in hdr:
-            object_name = hdr['OBJECT']
-            if object_name.upper() in self.all_names:
-                self.star_chooser.setCurrentText(object_name.upper())
+        star_name = response.lookup_std_star(hdr)
+        if star_name:
+            self.star_name = star_name
+            self.star_chooser.setCurrentText(self.star_name)
+        else:
+            warn_msg = "Could not load the name of the star.\nPlease select it from the dropdown menu."
+            WarningDialog(self, "No star name!", warn_msg)
 
         if self.ref_tab is not None:
             self.calculate_flux_in_bins()
@@ -340,12 +395,13 @@ class ResponseGUI(QtWidgets.QMainWindow):
 
 
     def set_star(self, text):
-        star_name = str(text).lower()
-        self.ref_tab = np.loadtxt(alfosc.path+'/calib/std/%s.dat' % star_name)
+        self.star_name = str(text)
+        self.ref_tab = np.loadtxt(response.path+'/calib/std/%s.dat' % self.star_name.lower())
         self.calculate_flux_in_bins()
         self.calculate_response_bins()
         self.update_plot()
 
+
     def calculate_flux_in_bins(self):
         if self.spectrum is None:
             WarningDialog(self, "No spectrum loaded!", "No spectral data has been loaded.")
@@ -394,10 +450,9 @@ class ResponseGUI(QtWidgets.QMainWindow):
         if self.spectrum is None:
             WarningDialog(self, "No spectrum loaded!", "No spectral data has been loaded.")
             return
-        if 'OBJECT' in self.spectrum.header:
-            object_name = self.spectrum.header['OBJECT']
-        else:
-            object_name = ''
+        object_name = instrument.get_object(self.spectrum.header)
+        if not object_name:
+            object_name = 'None'
 
         if self.data_line is None:
             self.data_line, = self.axes[0].plot(self.spectrum.wl, self.spectrum.data,
@@ -439,7 +494,6 @@ class ResponseGUI(QtWidgets.QMainWindow):
                 self.fit_line.set_data(self.spectrum.wl, self.response)
         self.canvas_points.draw()
 
-
     def fit_response(self):
         if self.resp_bins is None:
             WarningDialog(self, "No response data!", "No response data to fit.\nMake sure to load a spectrum and reference star data.")
@@ -453,7 +507,6 @@ class ResponseGUI(QtWidgets.QMainWindow):
         self.response = resp_fit(wl)
         self.update_points()
 
-
     def pick_points(self, event):
         x0 = event.mouseevent.xdata
         y0 = event.mouseevent.ydata
@@ -479,7 +532,6 @@ class ResponseGUI(QtWidgets.QMainWindow):
             return
         self.update_points()
 
-
     def create_menu(self):
         load_file_action = QtWidgets.QAction("Load Spectrum", self)
         load_file_action.setShortcut("ctrl+O")
@@ -493,6 +545,10 @@ class ResponseGUI(QtWidgets.QMainWindow):
         view_hdr_action.setShortcut("ctrl+shift+H")
         view_hdr_action.triggered.connect(self.display_header)
 
+        view_info_action = QtWidgets.QAction("Display Welcome Message", self)
+        view_info_action.setShortcut("ctrl+shift+I")
+        view_info_action.triggered.connect(lambda x: self.show_welcome(force=True))
+
         main_menu = self.menuBar()
         file_menu = main_menu.addMenu("File")
         file_menu.addAction(load_file_action)
@@ -500,7 +556,7 @@ class ResponseGUI(QtWidgets.QMainWindow):
 
         view_menu = main_menu.addMenu("View")
         view_menu.addAction(view_hdr_action)
-
+        view_menu.addAction(view_info_action)
 
     def display_header(self):
         if self.spectrum is not None:
@@ -577,6 +633,7 @@ class HeaderViewer(QtWidgets.QDialog):
         self.text_edit.setTextCursor(self.cursor)
 
 
+
 class WarningDialog(QtWidgets.QDialog):
     def __init__(self, parent, text, info=""):
         super(WarningDialog, self).__init__(parent)
