@@ -16,13 +16,17 @@ import warnings
 
 import spectres
 
-from pynot.alfosc import get_alfosc_header, create_pixel_array
-from pynot.functions import get_version_number, NN_mod_gaussian
-from pynot.scired import trim_overscan
+from pynot import instrument
+from pynot.functions import get_version_number, NN_mod_gaussian, get_pixtab_parameters, mad
 
 __version__ = get_version_number()
 
 
+class WavelengthError(Exception):
+    def __init__(self, message):
+        self.message = message
+
+
 def verify_arc_frame(arc_fname, dispaxis=2):
     """
     Return True if the arc lines fill the full detector plane, else return False.
@@ -32,9 +36,11 @@ def verify_arc_frame(arc_fname, dispaxis=2):
     `dispaxis` = 2 for vertical spectra, 1 for horizontal spectra.
     """
     arc2D = fits.getdata(arc_fname)
-    hdr = get_alfosc_header(arc_fname)
+    hdr = instrument.get_header(arc_fname)
     if 'DISPAXIS' in hdr:
         dispaxis = hdr['DISPAXIS']
+    elif instrument.get_dispaxis(hdr):
+        dispaxis = instrument.get_dispaxis(hdr)
 
     if dispaxis == 2:
         # Reorient image to have dispersion along x-axis:
@@ -142,7 +148,7 @@ def detect_borders(arc2D, kappa=20.):
     N_lines = np.zeros(arc2D.shape[0])
     for i, row in enumerate(arc2D):
         mad = np.median(np.abs(row-np.median(row)))
-        peaks, _ = find_peaks(row, prominence=kappa*mad)
+        peaks, _ = find_peaks(row - np.median(row), prominence=kappa*mad)
         N_lines[i] = len(peaks)
 
     mask = (N_lines >= np.median(N_lines)).nonzero()
@@ -196,7 +202,8 @@ def fit_2dwave_solution(pixtab2d, deg=5):
     return fit_table2d.T
 
 
-def apply_transform(img2D, pix, fit_table2d, ref_table, err2D=None, mask2D=None, header={}, order_wl=4, log=False, N_out=None, interpolate=True):
+def apply_transform(img2D, pix, fit_table2d, ref_table, err2D=None, mask2D=None, header={},
+                    order_wl=4, ref_type='vacuum', log=False, N_out=None, interpolate=True):
     """
     Apply 2D wavelength transformation to the input image
 
@@ -256,21 +263,50 @@ def apply_transform(img2D, pix, fit_table2d, ref_table, err2D=None, mask2D=None,
             msg.append("[WARNING] - Interpolation turned off!")
             msg.append("[WARNING] - N_out was given: %i" % N_out)
             msg.append("[WARNING] - Cannot change sampling without interpolating")
+            msg.append("[WARNING] - Going back to default: no interpolation and same dimension as input!")
 
     pix_in = pix
     cen = fit_table2d.shape[0]//2
     ref_wl = ref_table[:, 1]
     central_solution = Chebyshev.fit(fit_table2d[cen], ref_wl, deg=order_wl, domain=[pix_in.min(), pix_in.max()])
     wl_central = central_solution(pix_in)
+    if all(np.diff(wl_central) < 0):
+        # Wavelengths are decreasing: Flip arrays
+        flip_array = True
+    elif all(np.diff(wl_central) > 0):
+        # Wavelength are increasing: Nothing to do
+        flip_array = False
+    elif all(np.diff(wl_central) == 0):
+        # Wavelengths do not increase: WHAT?!
+        msg.append(" [ERROR]  - Wavelength array does not increase! Something went wrong.")
+        msg.append("          - Check the parameters `fit_window` and `order_wl`.")
+        exit_msg = "\n".join(msg)
+        raise WavelengthError(exit_msg)
+    else:
+        msg.append(" [ERROR]  - Wavelength array is not monotonic.")
+        msg.append("          - Check the parameters `fit_window` and `order_wl`.")
+        exit_msg = "\n".join(msg)
+        raise WavelengthError(exit_msg)
+
     wl_residuals = np.std(ref_wl - central_solution(fit_table2d[cen]))
-    msg.append("          - Residuals of wavelength solution: %.2f Å" % wl_residuals)
+    if wl_residuals > 5.:
+        msg.append("[WARNING] - Large residuals of wavelength solution: %.2f Å" % wl_residuals)
+        msg.append("[WARNING] - Try changing the fitting window of each arc line (par:`fit_window`)")
+    else:
+        msg.append("          - Residuals of wavelength solution: %.2f Å" % wl_residuals)
+
+    if ref_type == 'air':
+        ctype = 'AWAV'
+    else:
+        ctype = 'WAVE'
+
     if log:
         wl = np.logspace(np.log10(wl_central.min()), np.log10(wl_central.max()), N_out)
         hdr_tr = header.copy()
         hdr_tr['CRPIX1'] = 1
         hdr_tr['CDELT1'] = np.diff(np.log10(wl))[0]
         hdr_tr['CRVAL1'] = np.log10(wl[0])
-        hdr_tr['CTYPE1'] = 'LOGLAM  '
+        hdr_tr['CTYPE1'] = ctype+'-LOG'
         hdr_tr['CUNIT1'] = 'Angstrom'
         msg.append("          - Creating logarithmically sampled wavelength grid")
         msg.append("          - Sampling: %.3f  (logÅ/pix)" % np.diff(np.log10(wl))[0])
@@ -280,7 +316,7 @@ def apply_transform(img2D, pix, fit_table2d, ref_table, err2D=None, mask2D=None,
         hdr_tr['CRPIX1'] = 1
         hdr_tr['CDELT1'] = np.diff(wl)[0]
         hdr_tr['CRVAL1'] = wl[0]
-        hdr_tr['CTYPE1'] = 'LINEAR  '
+        hdr_tr['CTYPE1'] = ctype
         hdr_tr['CUNIT1'] = 'Angstrom'
         msg.append("          - Creating linearly sampled wavelength grid")
         msg.append("          - Sampling: %.3f  (Å/pix)" % np.diff(wl)[0])
@@ -296,36 +332,47 @@ def apply_transform(img2D, pix, fit_table2d, ref_table, err2D=None, mask2D=None,
         img2D_tr = np.zeros((img2D.shape[0], N_out))
         err2D_tr = np.zeros((img2D.shape[0], N_out))
         mask2D_tr = np.zeros((img2D.shape[0], N_out))
+        if err2D is None:
+            msg.append("[WARNING] - Interpolating data without errors!")
+        else:
+            msg.append("          - Interpolating data with errors")
+
         for i, row in enumerate(img2D):
             # - fit the chebyshev polynomium
             solution_row = Chebyshev.fit(fit_table2d[i], ref_wl, deg=order_wl, domain=[pix_in.min(), pix_in.max()])
             wl_row = solution_row(pix_in)
-            if np.diff(wl_row)[0] < 0:
+            if flip_array:
                 # Wavelengths are decreasing: Flip arrays
                 row = row[::-1]
                 wl_row = wl_row[::-1]
-                flip_array = True
-            else:
-                flip_array = False
 
             # -- interpolate the data onto the fixed wavelength grid
             if err2D is not None:
                 err_row = err2D[i]
                 if flip_array:
                     err_row = err_row[::-1]
-                interp_row, interp_err = spectres.spectres(wl, wl_row, row, spec_errs=err_row, verbose=False, fill=0.)
+
+                # interp_row, interp_err = spectres.spectres(wl, wl_row, row, spec_errs=err_row, verbose=False, fill=0.)
+                interp_row = np.interp(wl, wl_row, row, left=0., right=0.)
+                interp_err = np.interp(wl, wl_row, err_row, left=-1, right=-1)
                 err2D_tr[i] = interp_err
+                img2D_tr[i] = interp_row
             else:
-                interp_row = spectres.spectres(wl, wl_row, row, verbose=False, fill=0.)
+                # interp_row = spectres.spectres(wl, wl_row, row, verbose=False, fill=0.)
+                interp_row = np.interp(wl, wl_row, row, left=0., right=0.)
+                img2D_tr[i] = interp_row
 
             mask_row = mask2D[i]
             mask_int = np.interp(wl, wl_row, mask_row)
             mask2D_tr[i] = np.ceil(mask_int).astype(int)
-            img2D_tr[i] = interp_row
+
     else:
+        msg.append("          - No interpolation used!")
         img2D_tr = img2D
         err2D_tr = err2D
         mask2D_tr = mask2D
+    if np.sum(img2D_tr) == 0:
+        msg.append(" [ERROR]  - Something went wrong! All fluxes are 0!")
     output_msg = "\n".join(msg)
     return img2D_tr, err2D_tr, mask2D_tr, wl, hdr_tr, output_msg
 
@@ -354,17 +401,24 @@ def wavecal_1d(input_fname, pixtable_fname, *, output, order_wl=None, log=False,
 
     pixtable = np.loadtxt(pixtable_fname)
     msg.append("          - Loaded pixel table: %s" % pixtable_fname)
+    pixtab_pars, found_all = get_pixtab_parameters(pixtable_fname)
     if order_wl is None:
-        order_wl, found_in_file = get_order_from_file(pixtable_fname)
-        if found_in_file:
+        order_wl = pixtab_pars['order_wl']
+        if found_all:
             msg.append("          - Loaded polynomial order from file: %i" % order_wl)
         else:
             msg.append("          - Using default polynomial order: %i" % order_wl)
     else:
         msg.append("          - Using polynomial order: %i" % order_wl)
 
+    if pixtab_pars['ref_type'] == 'air':
+        ctype = 'AWAV'
+    else:
+        ctype = 'WAVE'
+
     if log:
         linearize = True
+        ctype = ctype + '-LOG'
 
     # Load input data:
     hdu_list = fits.open(input_fname)
@@ -382,6 +436,19 @@ def wavecal_1d(input_fname, pixtable_fname, *, output, order_wl=None, log=False,
         pix = tab['WAVE']
         flux1d = tab['FLUX']
         err1d = tab['ERR']
+        if 'OBJ_POS' in hdr:
+            aper_cen = hdr['OBJ_POS']
+            msg.append("          - Extraction aperture along slit at pixel no. %i" % aper_cen)
+            if pixtab_pars['loc'] != -1:
+                aperture_offset = np.abs(aper_cen - pixtab_pars['loc'])
+                if aperture_offset > 10:
+                    msg.append("[WARNING] - The wavelength solution may not be accurate!")
+                    msg.append("[WARNING] - The wavelength solution was calculated at position %i along the slit")
+                    msg.append("[WARNING] - but the spectrum was extracted at position %i along the slit.")
+        else:
+            aper_cen = -1
+            msg.append("[WARNING] - Extraction aperture not found.")
+            msg.append("[WARNING] - Double check if the wavelength solution is for the correct location along the slit!")
 
         if N_out is None:
             N_out = len(pix)
@@ -395,7 +462,9 @@ def wavecal_1d(input_fname, pixtable_fname, *, output, order_wl=None, log=False,
         res = np.std(pixtable[:, 1] - solution(pixtable[:, 0]))
         msg.append("          - Fitting wavelength solution with polynomium of order: %i" % order_wl)
         msg.append("          - Standard deviation of wavelength residuals: %.3f Å" % res)
+        msg.append("          - Setting header keyword: CTYPE1 = %s" % ctype)
         hdr['CUNIT1'] = 'Angstrom'
+        hdr['CTYPE1'] = ctype
         hdr['WAVERES'] = (np.round(res, 2), "RMS of wavelength residuals")
         if linearize:
             if log:
@@ -439,19 +508,20 @@ def wavecal_1d(input_fname, pixtable_fname, *, output, order_wl=None, log=False,
 
 # ============== PLOTTING =====================================================
 
-def plot_2d_pixtable(arc2D_sub, pix, pixtab2d, fit_table2d, filename=''):
+def plot_2d_pixtable(arc2D_sub, pix, pixtab2d, fit_table2d, arc_fname, filename=''):
     fig = plt.figure()
     ax = fig.add_subplot(111)
     mad = np.nanmedian(np.abs(arc2D_sub - np.nanmedian(arc2D_sub)))
     ax.imshow(arc2D_sub, origin='lower', extent=(pix.min(), pix.max(), 1, arc2D_sub.shape[0]),
-              vmin=-3*mad, vmax=10*mad, cmap=plt.cm.gray_r, aspect='auto')
-    for i in np.arange(0, pixtab2d.shape[0], 5):
+              vmin=-3*mad, vmax=15*mad, cmap=plt.cm.gray_r, aspect='auto')
+    for i in np.arange(0, pixtab2d.shape[0], 10):
         row = pixtab2d[i]
-        ax.plot(row, np.ones_like(row)*(i+1), ls='', color='Blue', marker='.', alpha=0.3)
+        ax.plot(row, np.ones_like(row)*(i+1), ls='', color='Blue', marker='.', alpha=0.1)
     for col in fit_table2d.T:
         ax.plot(col, np.arange(arc2D_sub.shape[0]), lw=1, color='r', alpha=0.5)
     ax.set_xlim(pix.min(), pix.max())
     ax.set_ylim(1, arc2D_sub.shape[0])
+    ax.set_title("Reference arc frame: %s" % arc_fname, fontsize=10)
     if filename:
         fig.savefig(filename)
 
@@ -472,40 +542,90 @@ def format_table2D_residuals(pixtab2d, fit_table2d, ref_table):
     wavelengths = ref_table[:, 1]
     resid2D = pixtab2d - fit_table2d
     for wl, resid_col, col in zip(wavelengths, resid2D.T, fit_table2d.T):
-        line_resid = np.std(resid_col)
+        line_resid = 1.48*mad(resid_col)
+        median_pix = np.median(col)
         delta_col = np.max(col) - np.min(col)
-        resid_log.append([wl, line_resid, delta_col])
+        resid_log.append([wl, median_pix, line_resid, delta_col])
     return resid_log
 
 # ============== MAIN ===========================================================
 
+def swap_axes_in_header(hdr):
+    cdelt1 = hdr.get('CDELT1')
+    cdelt2 = hdr.get('CDELT2')
+    cd_11 = hdr.get('CD1_1')
+    cd_22 = hdr.get('CD2_2')
+    cd_12 = hdr.get('CD1_2')
+    cd_21 = hdr.get('CD2_1')
+    crval1 = hdr.get('CRVAL1')
+    crval2 = hdr.get('CRVAL2')
+    crpix1 = hdr.get('CRPIX1')
+    crpix2 = hdr.get('CRPIX2')
+    ctype1 = hdr.get('CTYPE1')
+    ctype2 = hdr.get('CTYPE2')
+    cunit1 = hdr.get('CUNIT1')
+    cunit2 = hdr.get('CUNIT2')
+
+    if cdelt1:
+        hdr['CDELT2'] = cdelt1
+        hdr['CDELT1'] = cdelt2
+    if cd_11:
+        hdr['CD2_2'] = cd_11
+        hdr['CD1_1'] = cd_22
+    if cd_12:
+        hdr['CD1_2'] = cd_21
+        hdr['CD2_1'] = cd_12
+
+    hdr['CRVAL2'] = crval1
+    hdr['CRVAL1'] = crval2
+    hdr['CRPIX2'] = crpix1
+    hdr['CRPIX1'] = crpix2
+    hdr['CTYPE2'] = ctype1
+    hdr['CTYPE1'] = ctype2
+    hdr['CUNIT2'] = cunit1
+    hdr['CUNIT1'] = cunit2
+
+    binx = instrument.get_binx(hdr)
+    biny = instrument.get_biny(hdr)
+    hdr = instrument.set_binx(hdr, biny)
+    hdr = instrument.set_biny(hdr, binx)
+
+    return hdr
+
 def rectify(img_fname, arc_fname, pixtable_fname, output='', fig_dir='', order_bg=5, order_2d=5,
-            order_wl=4, log=False, N_out=None, interpolate=True, binning=1, dispaxis=2, fit_window=20,
-            plot=True, overwrite=True, verbose=False, overscan=50):
+            order_wl=4, log=False, N_out=None, interpolate=True, dispaxis=2, fit_window=20,
+            plot=True, overwrite=True, verbose=False, overscan=50, edge_kappa=10.):
 
     msg = list()
     arc2D = fits.getdata(arc_fname)
-    arc_hdr = get_alfosc_header(arc_fname)
-    arc2D, arc_hdr = trim_overscan(arc2D, arc_hdr, overscan=overscan)
     img2D = fits.getdata(img_fname)
     msg.append("          - Loaded image: %s" % img_fname)
     msg.append("          - Loaded reference arc image: %s" % arc_fname)
-    try:
-        err2D = fits.getdata(img_fname, 'ERR')
-        msg.append("          - Loaded error image")
-    except KeyError:
-        err2D = None
+    err2D = fits.getdata(img_fname, 'ERR')
+    msg.append("          - Loaded error image")
+
     try:
         mask2D = fits.getdata(img_fname, 'MASK')
         msg.append("          - Loaded mask image")
     except KeyError:
         mask2D = np.zeros_like(img2D)
-    hdr = fits.getheader(img_fname)
+    hdr = instrument.get_header(img_fname)
 
     ref_table = np.loadtxt(pixtable_fname)
+    pixtab_pars, found_all = get_pixtab_parameters(pixtable_fname)
     msg.append("          - Loaded reference pixel table: %s" % pixtable_fname)
+    ref_type = pixtab_pars['ref_type']
+    msg.append("          - Wavelength solution is in reference system: %s" % ref_type)
+    if found_all:
+        order_wl = pixtab_pars['order_wl']
+    else:
+        msg.append("[WARNING] - Not all parameters were loaded from the pixel table!")
+    msg.append("          - Polynomial order for wavelength as function of pixels : %i" % order_wl)
+
     if 'DISPAXIS' in hdr.keys():
         dispaxis = hdr['DISPAXIS']
+    elif instrument.get_dispaxis(hdr):
+        dispaxis = instrument.get_dispaxis(hdr)
 
     if dispaxis == 2:
         msg.append("          - Rotating frame to have dispersion along x-axis")
@@ -513,25 +633,13 @@ def rectify(img_fname, arc_fname, pixtable_fname, output='', fig_dir='', order_b
         arc2D = arc2D.T
         img2D = img2D.T
         mask2D = mask2D.T
-        if err2D is not None:
-            err2D = err2D.T
-        pix_in = create_pixel_array(hdr, dispaxis=2)
-
-        if 'DETYBIN' in hdr:
-            binning = hdr['DETYBIN']
-            hdr['DETYBIN'] = hdr['DETXBIN']
-            hdr['DETXBIN'] = binning
-        hdr['CDELT2'] = hdr['CDELT1']
-        hdr['CRVAL2'] = hdr['CRVAL1']
-        hdr['CRPIX2'] = hdr['CRPIX1']
-        hdr['CTYPE2'] = 'LINEAR'
-        hdr['CUNIT2'] = hdr['CUNIT1']
+        err2D = err2D.T
+        pix_in = instrument.create_pixel_array(hdr, axis=2)
+        hdr = swap_axes_in_header(hdr)
     else:
-        pix_in = create_pixel_array(hdr, dispaxis=1)
-        if 'DETXBIN' in hdr:
-            binning = hdr['DETXBIN']
+        pix_in = instrument.create_pixel_array(hdr, axis=1)
 
-    ilow, ihigh = detect_borders(arc2D)
+    ilow, ihigh = detect_borders(arc2D, kappa=edge_kappa)
     msg.append("          - Image shape: (%i, %i)" % arc2D.shape)
     msg.append("          - Detecting arc line borders: %i -- %i" % (ilow, ihigh))
     hdr['CRPIX2'] += ilow
@@ -557,28 +665,37 @@ def rectify(img_fname, arc_fname, pixtable_fname, output='', fig_dir='', order_b
     msg.append("          - Residuals of arc line positions relative to fitted 2D grid:")
     fit_residuals = format_table2D_residuals(pixtab2d, fit_table2d, ref_table)
 
-    msg.append("              Wavelength    Arc Residual   Max. Curvature")
-    for l0, line_residual, line_minmax in fit_residuals:
-        msg.append("              %10.3f    %-12.3f   %-14.3f" % (l0, line_residual, line_minmax))
+    msg.append("              Wavelength    Mean Position   Arc Residual   Max. Curvature")
+    for l0, med_line_pos, line_residual, line_minmax in fit_residuals:
+        msg.append("              %10.2f    %-13.2f   %-12.3f   %-14.3f" % (l0, med_line_pos, line_residual, line_minmax))
 
-    msg.append("          - Interpolating input image onto rectified wavelength solution")
-    transform_output = apply_transform(img2D, pix_in, fit_table2d, ref_table,
-                                       err2D=err2D, mask2D=mask2D, header=hdr, order_wl=order_wl,
-                                       log=log, N_out=N_out, interpolate=interpolate)
-    img2D_corr, err2D_corr, mask2D, wl, hdr_corr, trans_msg = transform_output
-    msg.append(trans_msg)
-    hdr.add_comment('PyNOT version %s' % __version__)
     if plot:
         plot_fname = os.path.join(fig_dir, 'PixTable2D.pdf')
-        plot_2d_pixtable(arc2D_sub, pix_in, pixtab2d, fit_table2d, filename=plot_fname)
+        plot_2d_pixtable(arc2D_sub, pix_in, pixtab2d, fit_table2d, arc_fname, filename=plot_fname)
         msg.append("          - Plotting fitted arc line positions in 2D frame")
         msg.append(" [OUTPUT] - Saving figure: %s" % plot_fname)
 
+    msg.append("          - Interpolating input image onto rectified wavelength solution")
+    try:
+        transform_output = apply_transform(img2D, pix_in, fit_table2d, ref_table,
+                                           err2D=err2D, mask2D=mask2D, header=hdr,
+                                           order_wl=order_wl, ref_type=ref_type,
+                                           log=log, N_out=N_out, interpolate=interpolate)
+        img2D_corr, err2D_corr, mask2D, wl, hdr_corr, trans_msg = transform_output
+        msg.append(trans_msg)
+
+    except WavelengthError as error:
+        msg.append(error.message)
+        output_str = "\n".join(msg)
+        print(output_str)
+        return "FATAL ERROR"
+
+    hdr.add_comment('PyNOT version %s' % __version__)
     if output:
         if output[-5:] != '.fits':
             output += '.fits'
     else:
-        object_name = hdr['OBJECT']
+        object_name = instrument.get_object(hdr)
         output = 'RECT2D_%s.fits' % (object_name)
 
     hdr_corr['DISPAXIS'] = 1
@@ -589,28 +706,31 @@ def rectify(img_fname, arc_fname, pixtable_fname, output='', fig_dir='', order_b
         if err2D_corr is not None:
             hdu['ERR'].data = err2D_corr
             err_hdr = hdu['ERR'].header
-            copy_keywords = ['CRPIX1', 'CRVAL1', 'CDELT1', 'CTYPE1', 'CUNIT1']
-            copy_keywords += ['CRPIX2', 'CRVAL2', 'CDELT2', 'CTYPE2', 'CUNIT2']
+            copy_keywords = ['CRPIX1', 'CRVAL1', 'CDELT1', 'CTYPE1', 'CUNIT1', 'CD1_1']
+            copy_keywords += ['CRPIX2', 'CRVAL2', 'CDELT2', 'CTYPE2', 'CUNIT2', 'CD2_2']
             for key in copy_keywords:
-                err_hdr[key] = hdr[key]
+                if key in hdr:
+                    err_hdr[key] = hdr[key]
             hdu['ERR'].header = err_hdr
         if 'MASK' in hdu:
             hdu['MASK'].data = mask2D
             mask_hdr = hdu['MASK'].header
-            copy_keywords = ['CRPIX1', 'CRVAL1', 'CDELT1', 'CTYPE1', 'CUNIT1']
-            copy_keywords += ['CRPIX2', 'CRVAL2', 'CDELT2', 'CTYPE2', 'CUNIT2']
+            copy_keywords = ['CRPIX1', 'CRVAL1', 'CDELT1', 'CTYPE1', 'CUNIT1', 'CD1_1']
+            copy_keywords += ['CRPIX2', 'CRVAL2', 'CDELT2', 'CTYPE2', 'CUNIT2', 'CD2_2']
             for key in copy_keywords:
-                mask_hdr[key] = hdr[key]
+                if key in hdr:
+                    mask_hdr[key] = hdr[key]
             hdu['MASK'].header = mask_hdr
         else:
             mask_hdr = fits.Header()
             mask_hdr.add_comment("2 = Good Pixels")
             mask_hdr.add_comment("1 = Cosmic Ray Hits")
             mask_hdr['AUTHOR'] = 'PyNOT version %s' % __version__
-            copy_keywords = ['CRPIX1', 'CRVAL1', 'CDELT1', 'CTYPE1', 'CUNIT1']
-            copy_keywords += ['CRPIX2', 'CRVAL2', 'CDELT2', 'CTYPE2', 'CUNIT2']
+            copy_keywords = ['CRPIX1', 'CRVAL1', 'CDELT1', 'CTYPE1', 'CUNIT1', 'CD1_1']
+            copy_keywords += ['CRPIX2', 'CRVAL2', 'CDELT2', 'CTYPE2', 'CUNIT2', 'CD2_2']
             for key in copy_keywords:
-                mask_hdr[key] = hdr[key]
+                if key in hdr:
+                    mask_hdr[key] = hdr[key]
             mask_ext = fits.ImageHDU(mask2D, header=mask_hdr, name='MASK')
             hdu.append(mask_ext)
         hdu.writeto(output, overwrite=overwrite)
