@@ -74,7 +74,7 @@ def match_catalogs(coords, refs):
     return np.array(matched)
 
 
-def get_gaia_catalog(ra, dec, radius=4., limit=200, catalog_fname='', database='edr3'):
+def get_gaia_catalog(ra, dec, radius=4., limit=500, catalog_fname='', database='edr3'):
     """
     ra and dec: units of degrees
     radius: units of arcmin
@@ -94,7 +94,7 @@ def get_gaia_catalog(ra, dec, radius=4., limit=200, catalog_fname='', database='
 
 
 
-def correct_wcs(img_fname, sep_fname, output_fname='', fig_fname='', max_num=60, min_num=10, G_lim=15, q_lim=0.8, kappa=3):
+def correct_wcs(img_fname, sep_fname, output='', fig_fname='', max_num=60, min_num=10, G_lim=15, q_lim=0.8, kappa=3):
     """
     WCS calibration using Gaia
 
@@ -109,7 +109,7 @@ def correct_wcs(img_fname, sep_fname, output_fname='', fig_fname='', max_num=60,
     sep_fname : string
         Filename of the source extraction table (_phot.fits)
 
-    output_fname : string
+    output : string
         Filename of the WCS corrected image. Autogenerated by default.
 
     fig_fname : string
@@ -147,10 +147,12 @@ def correct_wcs(img_fname, sep_fname, output_fname='', fig_fname='', max_num=60,
     # Prepare output filenames:
     base, ext = os.path.splitext(os.path.basename(img_fname))
     dirname = os.path.dirname(img_fname)
-    if output_fname == '':
+    if output == '':
         output_fname = img_fname
         # output_fname = base + '_wcs' + ext
         # output_fname = os.path.join(dirname, output_fname)
+    else:
+        output_fname = output
 
     if fig_fname == '':
         fig_fname = 'wcs_solution_' + base + '.pdf'
@@ -235,7 +237,14 @@ def correct_wcs(img_fname, sep_fname, output_fname='', fig_fname='', max_num=60,
     wcs_keys = update_WCS(matched_sep, matched_ref, crval, CD)
     hdr.update(wcs_keys)
     hdr.add_comment("PyNOT: WCS calibration using Gaia %s" % gaia_dr)
-    hdr['RADESYSa'] = hdr['RADECSYS']
+    if 'RADESYSa' not in hdr:
+        if 'RADECSYS' in hdr:
+            hdr['RADESYSa'] = hdr['RADECSYS']
+        else:
+            msg.append("  [ERROR] - Could not find header keyword: RADECSYS or RADECSYSa")
+            output_msg = "\n".join(msg)
+            return output_msg
+
     if 'RADECSYS' in hdr:
         hdr.pop('RADECSYS')
     msg.append("          - Updating WCS information")
